from psychopy import core

from bcipy.tasks.task import Task
from bcipy.display.rsvp.mode.copy_phrase import CopyPhraseDisplay
from bcipy.feedback.visual.visual_feedback import VisualFeedback
from bcipy.helpers.triggers import _write_triggers_from_sequence_copy_phrase
from bcipy.helpers.save import _save_session_related_data
from bcipy.helpers.signal_model import CopyPhraseWrapper
from bcipy.helpers.task import (
    fake_copy_phrase_decision, alphabet, process_data_for_decision,
    trial_complete_message, get_user_input)


class RSVPCopyPhraseTask(Task):
    """RSVP Copy Phrase Task.

    Initializes and runs all needed code for executing a copy phrase task. A
        phrase is set in parameters and necessary objects (daq, display) are
        passed to this function. Certain Wrappers and Task Specific objects are
        executed here.

    Parameters
    ----------
        win : object,
            display window to present visual stimuli.
        daq : object,
            data acquisition object initialized for the desired protocol
        parameters : dict,
            configuration details regarding the experiment. See parameters.json
        file_save : str,
            path location of where to save data from the session
        signal_model : loaded pickle file,
            trained signal model.
        language_model: object,
            trained language model.
        fake : boolean, optional
            boolean to indicate whether this is a fake session or not.
    Returns
    -------
        file_save : str,
            path location of where to save data from the session
    """

    TASK_NAME = 'RSVP Copy Phrase Task'

    def __init__(
            self, win, daq, parameters, file_save, signal_model, language_model, fake):
        super(RSVPCopyPhraseTask, self).__init__()

        self.window = win
        self.frame_rate = self.window.getActualFrameRate()
        self.parameters = parameters
        self.daq = daq
        self.static_clock = core.StaticPeriod(screenHz=self.frame_rate)
        self.experiment_clock = core.Clock()
        self.buffer_val = parameters['task_buffer_len']
        self.alp = alphabet(parameters)
        self.rsvp = _init_copy_phrase_display(
            self.parameters, self.window, self.daq,
            self.static_clock, self.experiment_clock)
        self.file_save = file_save

        trigger_save_location = f"{self.file_save}/{parameters['trigger_file_name']}"
        self.trigger_file = open(trigger_save_location, 'w')
        self.session_save_location = f"{self.file_save}/{parameters['session_file_name']}"

        self.wait_screen_message = parameters['wait_screen_message']
        self.wait_screen_message_color = parameters[
            'wait_screen_message_color']

        self.stim_number = parameters['stim_number']
        self.stim_length = parameters['stim_length']
        self.time_cross = parameters['time_cross']
        self.time_target = parameters['time_target']
        self.time_flash = parameters['time_flash']
        self.timing = [self.time_target,
                       self.time_cross,
                       self.time_flash]

        self.color = [parameters['target_letter_color'],
                      parameters['fixation_color'],
                      parameters['stim_color']]

        self.task_info_color = parameters['task_color']

        self.stimuli_height = parameters['stim_height']

        self.is_txt_stim = parameters['is_txt_stim']
        self.eeg_buffer = parameters['eeg_buffer_len']
        self.copy_phrase = parameters['task_text']
        self.spelled_letters_count = int(
            parameters['spelled_letters_count'])
        if self.spelled_letters_count > len(self.copy_phrase):
            self.logger.debug('Already spelled letters exceeds phrase length.')
            self.spelled_letters_count = 0

        self.max_seq_length = parameters['max_seq_len']
        self.max_seconds = parameters['max_minutes'] * 60  # convert to seconds
        self.max_seq_per_trial = parameters['max_seq_per_trial']
        self.fake = fake
        self.language_model = language_model
        self.signal_model = signal_model
        self.down_sample_rate = parameters['down_sampling_rate']

        self.filter_low = self.parameters['filter_low_pass']
        self.filter_high = self.parameters['filter_high_pass']
        self.fitler_order = self.parameters['filter_order']
        self.notch_filter_frequency = self.parameters['notch_filter_frequency']

        self.min_num_seq = parameters['min_seq_len']
        self.collection_window_len = parameters['collection_window_after_trial_length']

        self.static_offset = parameters['static_trigger_offset']
        self.show_feedback = parameters['show_feedback']
        self.feedback_color = parameters['feedback_message_color']

        if self.show_feedback:
            self.feedback = VisualFeedback(
                self.window, self.parameters, self.experiment_clock)

    def execute(self):
        self.logger.debug('Starting Copy Phrase Task!')

        # already correctly spelled letters
        text_task = str(self.copy_phrase[0:self.spelled_letters_count])
        task_list = [(str(self.copy_phrase),
                      str(self.copy_phrase[0:self.spelled_letters_count]))]

        # Try Initializing Copy Phrase Wrapper:
        copy_phrase_task = CopyPhraseWrapper(
            self.min_num_seq,
            self.max_seq_per_trial,
            signal_model=self.signal_model,
            fs=self.daq.device_info.fs,
            k=2,
            alp=self.alp,
            task_list=task_list,
            lmodel=self.language_model,
            is_txt_stim=self.is_txt_stim,
            device_name=self.daq.device_info.name,
            device_channels=self.daq.device_info.channels,
            stimuli_timing=[self.time_cross, self.time_flash],
            decision_threshold=self.parameters['decision_threshold'],
            backspace_prob=self.parameters['lm_backspace_prob'],
            backspace_always_shown=self.parameters['backspace_always_shown'],
            filter_high=self.filter_high,
            filter_low=self.filter_low,
            filter_order=self.fitler_order,
            notch_filter_frequency=self.notch_filter_frequency)

        # Set new epoch (whether to present a new epoch),
        #   run (whether to cont. session),
        #   sequence counter (how many seq have occured).
        #   epoch counter and index
        #   (what epoch, and how many sequences within it)
        new_epoch = True
        run = True
        seq_counter = 0
        epoch_counter = 0
        epoch_index = 0

        # Init session data and save before beginning
        data = {
            'session': self.file_save,
            'session_type': 'Copy Phrase',
            'paradigm': 'RSVP',
            'epochs': {},
            'total_time_spent': self.experiment_clock.getTime(),
            'total_number_epochs': 0,
        }

        # Save session data
        _save_session_related_data(self.session_save_location, data)

        # check user input to make sure we should be going
        if not get_user_input(self.rsvp, self.wait_screen_message,
                              self.wait_screen_message_color,
                              first_run=True):
            run = False

        # Start the Session!
        while run:

            # check user input to make sure we should be going
            if not get_user_input(self.rsvp, self.wait_screen_message,
                                  self.wait_screen_message_color):
                break

            if self.copy_phrase[0:len(text_task)] == text_task:
                # if correctly spelled so far, get the next unspelled letter.
                target_letter = self.copy_phrase[len(text_task)]
            else:
                # otherwise target is the backspace char.
                target_letter = '<'

            # Get sequence information
            if new_epoch:

                # Init an epoch, getting initial stimuli
                new_epoch, sti = copy_phrase_task.initialize_epoch()
                ele_sti = sti[0]
                timing_sti = sti[1]
                color_sti = sti[2]

                # Increase epoch number and reset epoch index
                epoch_counter += 1
                data['epochs'][epoch_counter] = {}
                epoch_index = 0
            else:
                epoch_index += 1

            # Update task state and reset the static
            self.rsvp.update_task_state(text=text_task, color_list=['white'])
            self.rsvp.draw_static()
            self.window.flip()

            # Setup the new Stimuli
            self.rsvp.stimuli_sequence = ele_sti[0]
            if self.is_txt_stim:
                self.rsvp.stimuli_colors = color_sti[0]
            self.rsvp.stimuli_timing = timing_sti[0]

            # Pause for a time
            core.wait(self.buffer_val)

            # Do the self.RSVP sequence!
            sequence_timing = self.rsvp.do_sequence()

            # Write triggers to file
            _write_triggers_from_sequence_copy_phrase(
                sequence_timing,
                self.trigger_file,
                self.copy_phrase,
                text_task)

            core.wait(self.buffer_val)

            # Delete calibration
            if seq_counter == 0:
                del sequence_timing[0]

            # reshape the data and triggers as needed for later modules
            raw_data, triggers, target_info = \
                process_data_for_decision(
                    sequence_timing,
                    self.daq,
                    self.window,
                    self.parameters,
                    self.rsvp.first_stim_time,
                    self.static_offset)

            # Uncomment this to turn off fake decisions, but use fake data.
            # self.fake = False
            if self.fake:
                # Construct Data Record
                data['epochs'][epoch_counter][epoch_index] = {
                    'stimuli': ele_sti,
                    'eeg_len': len(raw_data),
                    'timing_sti': timing_sti,
                    'triggers': triggers,
                    'target_info': target_info,
                    'target_letter': target_letter,
                    'current_text': text_task,
                    'copy_phrase': self.copy_phrase}

                # Evaluate this sequence
                (target_letter, text_task, run) = \
                    fake_copy_phrase_decision(self.copy_phrase,
                                              target_letter,
                                              text_task)

                # here we assume, in fake mode, all sequences result in a selection.
                last_selection = text_task[-1]
                new_epoch = True
                # Update next state for this record
                data['epochs'][
                    epoch_counter][
                    epoch_index][
                    'next_display_state'] = \
                    text_task

            else:
                # Evaluate this sequence, returning whether to gen a new
                #  epoch (seq) or stimuli to present
                new_epoch, sti = \
                    copy_phrase_task.evaluate_sequence(raw_data, triggers,
                                                       target_info, self.collection_window_len)

                # Construct Data Record
                data['epochs'][epoch_counter][epoch_index] = {
                    'stimuli': ele_sti,
                    'eeg_len': len(raw_data),
                    'timing_sti': timing_sti,
                    'triggers': triggers,
                    'target_info': target_info,
                    'current_text': text_task,
                    'copy_phrase': self.copy_phrase,
                    'next_display_state':
                        copy_phrase_task.decision_maker.displayed_state,
                    'lm_evidence': copy_phrase_task
                        .conjugator
                        .evidence_history['LM'][0]
                        .tolist(),
                    'eeg_evidence': copy_phrase_task
                        .conjugator
                        .evidence_history['ERP'][-1]
                        .tolist(),
                    'likelihood': copy_phrase_task
                        .conjugator.likelihood.tolist()
                }

                # If new_epoch is False, get the stimuli info returned
                if not new_epoch:
                    ele_sti = sti[0]
                    timing_sti = sti[1]
                    color_sti = sti[2]

                # Get the current task text from the decision maker
                text_task = copy_phrase_task.decision_maker.displayed_state
                last_selection = copy_phrase_task.decision_maker.last_selection

            # if a letter was selected and feedback enabled, show the chosen letter
            if new_epoch and self.show_feedback:
                self.feedback.administer(
                    last_selection,
                    message='Selected:',
                    line_color=self.feedback_color,
                    fill_color=self.feedback_color)

            # Update time spent and save data
            data['total_time_spent'] = self.experiment_clock.getTime()
            data['total_number_epochs'] = epoch_counter
            _save_session_related_data(self.session_save_location, data)

            # Decide whether to keep the task going
            max_tries_exceeded = seq_counter >= self.max_seq_length
            max_time_exceeded = data['total_time_spent'] >= self.max_seconds
            if (text_task == self.copy_phrase or max_tries_exceeded or
                    max_time_exceeded):
                if max_tries_exceeded:
                    self.logger.debug('COPYPHRASE ERROR: Max tries exceeded. To allow for more tries '
                                      'adjust the max_seq_len parameter.')
                if max_time_exceeded:
                    self.logger.debug('COPYPHRASE ERROR: Max time exceeded. To allow for more time '
                                      'adjust the max_minutes parameter.')
                run = False

            # Increment sequence counter
            seq_counter += 1

        # Update task state and reset the static
        self.rsvp.update_task_state(text=text_task, color_list=['white'])

        # Say Goodbye!
        self.rsvp.text = trial_complete_message(self.window, self.parameters)
        self.rsvp.draw_static()
        self.window.flip()

        # Give the system time to process
        core.wait(self.buffer_val)

        if self.daq.is_calibrated:
            _write_triggers_from_sequence_copy_phrase(
                ['offset', self.daq.offset], self.trigger_file,
                self.copy_phrase, text_task, offset=True)

        # Close the trigger file for this session
        self.trigger_file.close()

        # Wait some time before exiting so there is trailing eeg data saved
        core.wait(self.eeg_buffer)

        return self.file_save

    def name(self):
        return self.TASK_NAME


def _init_copy_phrase_display(
        parameters, win, daq, static_clock, experiment_clock):
    return CopyPhraseDisplay(
        win,
        static_clock,
        experiment_clock,
        daq.marker_writer,
        info_text=parameters['info_text'],
        static_task_text=parameters['task_text'],
        task_text='****',
        info_color=parameters['info_color'],
        info_pos=(parameters['text_pos_x'],
                  parameters['text_pos_y']),
        info_height=parameters['info_height'],
        info_font=parameters['info_font'],
        task_color=[parameters['task_color']],
        task_font=parameters['task_font'],
        task_height=parameters['task_height'],
        stim_font=parameters['stim_font'],
        stim_pos=(parameters['stim_pos_x'],
                 parameters['stim_pos_y']),
        stim_height=parameters['stim_height'],
        stim_sequence=['a'] * 10,
        stim_colors=[parameters['stim_color']] * 10,
        stim_timing=[3] * 10,
        is_txt_stim=parameters['is_txt_stim'],
        trigger_type=parameters['trigger_type'],
        space_char=parameters['stim_space_char'])



import unittest
import numpy as np

from bcipy.helpers.signal_model import CopyPhraseWrapper
from bcipy.helpers.task import alphabet


class TestSignalModelRelated(unittest.TestCase):
    """Test CopyPhraseWrapper"""

    def test_valid_letters(self):
        alp = alphabet()
        cp = CopyPhraseWrapper(
            min_num_seq=1,
            max_num_seq=50,
            signal_model=None,
            fs=25,
            k=2,
            alp=alp,
            task_list=[('HELLO_WORLD', 'HE')],
            is_txt_stim=True,
            device_name='LSL',
            evidence_names=['LM', 'ERP'],
            device_channels=[
                'ch1', 'ch2', 'ch3', 'ch4', 'ch5', 'ch6', 'ch7', 'ch8', 'ch9',
                'ch10', 'ch11', 'ch12', 'ch13', 'ch14', 'ch15', 'ch16', 'TRG'
            ],
            stimuli_timing=[0.5, 0.25])

        triggers = [('+', 0.0), ('H', 0.5670222830376588),
                    ('D', 0.8171830819919705), ('J', 1.0843321380089037),
                    ('B', 1.3329724550130777), ('C', 1.5825864360085689),
                    ('A', 1.833380013005808), ('F', 2.083211077027954),
                    ('G', 2.333359022042714), ('I', 2.583265081048012),
                    ('E', 2.833274284028448)]
        target_info = [
            'nontarget', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget'
        ]

        letters, timings, labels = cp.letter_info(triggers, target_info)
        expected_letters = ['H', 'D', 'J', 'B', 'C', 'A', 'F', 'G', 'I', 'E']
        expected_time = [
            0.5670222830376588, 0.8171830819919705, 1.0843321380089037,
            1.3329724550130777, 1.5825864360085689, 1.833380013005808,
            2.083211077027954, 2.333359022042714, 2.583265081048012,
            2.833274284028448
        ]
        self.assertEqual(expected_letters, letters)
        self.assertEqual(expected_time, timings)
        self.assertEqual(len(letters), len(labels))

        triggers = [('calibration_trigger', 0.0), ('+', 0.1), ('H', 0.5670222830376588),
                    ('D', 0.8171830819919705), ('J', 1.0843321380089037),
                    ('B', 1.3329724550130777), ('C', 1.5825864360085689),
                    ('A', 1.833380013005808), ('F', 2.083211077027954),
                    ('G', 2.333359022042714), ('I', 2.583265081048012),
                    ('E', 2.833274284028448)]
        target_info = [
            'calib', 'fixation', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget', 'nontarget', 'nontarget', 'nontarget', 'nontarget',
            'nontarget'
        ]
        letters, timings, labels = cp.letter_info(triggers, target_info)
        self.assertEqual(expected_letters, letters)
        self.assertEqual(expected_time, timings)
        self.assertEqual(['nontarget'] * (len(letters)), labels)

        # Test it throws an exception when letter is outside alphabet
        with self.assertRaises(Exception):
            cp.letter_info([('A', 0.0), ('*', 1.0)], ['nontarget', 'nontarget'])


import os
import numpy as np
import matplotlib.pyplot as plt
from bcipy.helpers.load import load_csv_data, read_data_csv

import logging
log = logging.getLogger(__name__)


def generate_offline_analysis_screen(
        x,
        y,
        model=None,
        folder=None,
        plot_lik_dens=True,
        save_figure=True,
        down_sample_rate=2,
        fs=300,
        plot_x_ticks=8,
        plot_average=False,
        show_figure=False,
        channel_names = None) -> None:
    """ Offline Analysis Screen.

    Generates the information figure following the offlineAnalysis.
    The figure has multiple tabs containing the average ERP plots

    PARAMETERS
    ----------

    x(ndarray[float]): C x N x k data array
    y(ndarray[int]): N x k observation (class) array
        N is number of samples k is dimensionality of features
        C is number of channels
    model(): trained model for data
    folder(str): Folder of the data
    plot_lik_dens: boolean: whether or not to plot likelihood densities
    save_figures: boolean: whether or not to save the plots as PDF
    down_sample_rate: downsampling rate applied to signal (if any)
    fs (sampling_rate): original sampling rate of the signal
    plot_x_ticks: number of ticks desired on the ERP plot
    plot_average: boolean: whether or not to average over all channels
    show_figure: boolean: whether or not to show the figures generated
    channel_names: dict of channel names keyed by their position.
    """
    channel_names = channel_names or {}
    classes = np.unique(y)

    means = [np.squeeze(np.mean(x[:, np.where(y == i), :], 2)) for i in classes]

    fig = plt.figure(figsize=(20, 10))
    ax1 = fig.add_subplot(211)
    ax2 = fig.add_subplot(212)

    if plot_average:
        # find the mean across rows for non target and target
        non_target_mean = np.mean(means[0], axis=0)
        target_mean = np.mean(means[1], axis=0)

        # plot the means
        ax1.plot(non_target_mean)
        ax2.plot(target_mean)

    else:
        count = 0
        # Loop through all the channels and plot each on the non target/target subplots
        while count < means[0].shape[0]:
            lbl = channel_names.get(count, count)
            ax1.plot(means[0][count, :], label=lbl)
            ax2.plot(means[1][count, :], label=lbl)
            count += 1
        ax1.legend(loc='upper left',  prop={'size': 8})
        ax2.legend(loc='upper left', prop={'size': 8})

    # data points
    data_length = len(means[0][1, :])

    # generate appropriate data labels for the figure
    lower = 0

    # find the upper length of data and convert to seconds
    upper = data_length * down_sample_rate / fs * 1000

    # make the labels
    labels = [round(lower + x * (upper - lower) / (plot_x_ticks - 1)) for x in
              range(plot_x_ticks)]

    # make sure it starts at zero
    labels.insert(0, 0)

    # set the labels
    ax1.set_xticklabels(labels)
    ax2.set_xticklabels(labels)

    # Set common labels
    fig.text(0.5, 0.04, 'Time (Seconds)', ha='center', va='center')
    fig.text(0.06, 0.5, '$\mu V$', ha='center', va='center',
             rotation='vertical')

    ax1.set_title('Non-target ERP')
    ax2.set_title('Target ERP')

    if save_figure:
        fig.savefig(os.path.join(folder, 'mean_erp.pdf'), bbox_inches='tight', format='pdf')

    if plot_lik_dens:
        fig, ax = plt.subplots()
        x_plot = np.linspace(np.min(model.line_el[-2]), np.max(model.line_el[-2]), 1000)[:, np.newaxis]
        ax.plot(model.line_el[2][y == 0], -0.005 - 0.01 * np.random.random(model.line_el[2][y == 0].shape[0]), 'ro', label='class(-)')
        ax.plot(model.line_el[2][y == 1], -0.005 - 0.01 * np.random.random(model.line_el[2][y == 1].shape[0]), 'go', label='class(+)')
        for idx in range(len(model.pipeline[2].list_den_est)):
            log_dens = model.pipeline[2].list_den_est[idx].score_samples(x_plot)
            ax.plot(x_plot[:, 0], np.exp(log_dens),'r-' * (idx == 0) + 'g--' * (idx == 1), linewidth=2.0)

        ax.legend(loc='upper right')
        plt.title('Likelihoods Given the Labels')
        plt.ylabel('p(e|l)')
        plt.xlabel('scores')

        if save_figure:
            fig.savefig(os.path.join(folder, 'lik_dens.pdf'), bbox_inches='tight', format='pdf')

    if show_figure:
        plt.show()


def visualize_csv_eeg_triggers(trigger_col=None):
    """Visualize CSV EEG Triggers.

    This function is used to load in CSV data and visualize device generated triggers.

    Input:
        trigger_col(int)(optional): Column location of triggers in csv file.
            It defaults to the last column.

    Output:
        Figure of Triggers
    """
    # Load in CSV
    file_name = load_csv_data()
    raw_data, stamp_time, channels, type_amp, fs = read_data_csv(file_name)

    # Pull out the triggers
    if not trigger_col:
        triggers = raw_data[-1]
    else:
        triggers = raw_data[trigger_col]

    # Plot the triggers
    plt.plot(triggers)

    # Add some titles and labels to the figure
    plt.title('Trigger Signal')
    plt.ylabel('Trigger Value')
    plt.xlabel('Samples')

    log.debug('Press Ctrl + C to exit!')
    # Show us the figure! Depending on your OS / IDE this may not close when
    #  The window is closed, see the message above
    plt.show()


if __name__ == '__main__':
    import pickle

    # load some x, y data from test files
    x = pickle.load(open('bcipy/helpers/tests/resources/mock_x_generate_erp.pkl', 'rb'))
    y = pickle.load(open('bcipy/helpers/tests/resources/mock_y_generate_erp.pkl', 'rb'))

    names = {
        0: 'P3',
        1: 'C3',
        2: 'F3',
        3: 'Fz',
        4: 'F4',
        5: 'C4',
        6: 'P4',
        7: 'Cz',
        8: 'A1',
        9: 'Fp1',
        10: 'Fp2',
        11: 'T3',
        12: 'T5',
        13: 'O1',
        14: 'O2',
        15: 'F7',
        16: 'F8',
        17: 'A2',
        18: 'T6',
        19: 'T4'
    }
    # generate the offline analysis screen. show figure at the end
    generate_offline_analysis_screen(
        x,
        y,
        folder='bcipy',
        plot_lik_dens=False,
        save_figure=False,
        show_figure=True,
        channel_names=names)




class InsufficientDataException(Exception):
    """Insufficient Data Exception.

    Thrown when data requirments to execute task are violated.
    """
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors


class TaskRegistryException(Exception):
    """Task Registry Exception.

    Thrown when task type unregistered.
    """
    def __init__(self, message, errors=None):
        super().__init__(message)
        self.errors = errors



# pylint: disable=no-self-use
"""Tests for the LabStreamingLayer driver."""

import unittest
import pytest
from bcipy.acquisition.datastream.lsl_server import LslDataServer
from bcipy.acquisition.datastream.server import await_start
from bcipy.acquisition.datastream import generator
from bcipy.acquisition.protocols.lsl.lsl_device import LslDevice


class TestLslDevice(unittest.TestCase):
    """Test the LslDevice"""

    def __init__(self, *args, **kwargs):
        super(TestLslDevice, self).__init__(*args, **kwargs)
        self.channels = ['C3', 'C4', 'Cz', 'FPz', 'POz', 'CPz', 'O1', 'O2']
        self.channel_count = len(self.channels)
        # pylint: disable=invalid-name
        self.hz = 100

    @property
    def include_meta(self):
        """Whether or not to include metadata"""
        raise Exception("Must be implemented in subclass")

    def setUp(self):
        """Run before each test."""
        self.server = LslDataServer(params={'name': 'LSL',
                                            'channels': self.channels,
                                            'hz': self.hz},
                                    generator=generator.random_data(
                                        channel_count=self.channel_count),
                                    include_meta=self.include_meta,
                                    add_markers=True)
        await_start(self.server)

    def tearDown(self):
        """Run after each test."""
        self.server.stop()
        self.server = None


class TestLslWithoutMetadata(TestLslDevice):
    """LSL Device tests in which the server does not provide metadata."""

    @property
    def include_meta(self):
        return False

    def test_incorrect_number_of_channels(self):
        """A list of channels with len that does not match channel_count should
        raise an exception."""

        device = LslDevice(connection_params={}, fs=self.hz,
                           channels=['ch1', 'ch2'])
        self.assertEqual(len(device.channels), 2)
        self.assertNotEqual(len(device.channels), self.channel_count)
        device.connect()

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_incorrect_frequency(self):
        """Provided fs should match sample rate read from device"""
        device = LslDevice(connection_params={},
                           channels=self.channels, fs=300)
        self.assertEqual(device.fs, 300)
        device.connect()

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_frequency_init(self):
        """fs should be initialized from device metadata if not provided"""

        device = LslDevice(connection_params={},
                           channels=self.channels + ["TRG"], fs=None)
        device.connect()
        device.acquisition_init()

        self.assertEqual(device.fs, self.hz)

    def test_connect(self):
        """Should require a connect call before initialization."""
        device = LslDevice(connection_params={},
                           channels=[], fs=None)

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_read_data(self):
        """Should produce a valid data record."""

        device = LslDevice(connection_params={},
                           channels=self.channels + ["TRG"], fs=self.hz)

        device.connect()
        device.acquisition_init()
        data = device.read_data()

        self.assertTrue(len(data) > 0)
        self.assertEqual(len(data), len(device.channels))

        for channel in data[0:-1]:
            self.assertTrue(isinstance(channel, float))


class TestLslWithMetadata(TestLslDevice):
    """LSL Device tests in which the server provides metadata."""

    @property
    def include_meta(self):
        return True

    def test_mismatched_channel_names(self):
        """Provided channel names should match device information."""
        channels = ['ch' + str(i) for i in range(self.channel_count)]
        device = LslDevice(connection_params={}, fs=self.hz, channels=channels)

        self.assertEqual(len(device.channels), self.channel_count)
        device.connect()

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_channel_init(self):
        """Channels should be initialized from device metadata if not
         provided."""
        device = LslDevice(connection_params={}, fs=self.hz,
                           channels=[])
        self.assertEqual(len(device.channels), 0)
        device.connect()
        device.acquisition_init()
        self.assertEqual(len(device.channels), self.channel_count + 1,
                         "Should have added a TRG channel")
        self.assertEqual(device.channels[0:-1], self.channels)

    def test_frequency_init(self):
        """fs should be initialized from device metadata if not provided"""

        device = LslDevice(connection_params={}, fs=None)
        device.connect()
        device.acquisition_init()

        self.assertEqual(device.fs, self.hz)

    def test_connect(self):
        """Should require a connect call before initialization."""
        device = LslDevice(connection_params={},
                           channels=[], fs=None)

        with pytest.raises(Exception):
            device.acquisition_init()

    def test_read_data(self):
        """Should produce a valid data record."""

        device = LslDevice(connection_params={}, fs=self.hz)

        device.connect()
        device.acquisition_init()
        data = device.read_data()

        self.assertTrue(len(data) > 0)
        self.assertEqual(len(data), len(device.channels))
        for channel in data[0:-1]:
            self.assertTrue(isinstance(channel, float))


if __name__ == '__main__':
    unittest.main()



import wx
import wx.lib.scrolledpanel as scrolled
import json
import logging
from typing import Callable, Dict, Tuple
from collections import namedtuple
from os import sep

log = logging.getLogger(__name__)
JSON_INDENT = 2


def font(size: int=14, font_family: wx.Font=wx.FONTFAMILY_SWISS) -> wx.Font:
    """Create a Font object with the given parameters."""
    return wx.Font(size, font_family, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_LIGHT)


def static_text_control(parent,
                        label: str,
                        color: str = 'black',
                        size: int = 14,
                        font_family: wx.Font=wx.FONTFAMILY_SWISS) -> wx.StaticText:
    """Creates a static text control with the given font parameters. Useful for
    creating labels and help components."""

    static_text = wx.StaticText(parent, label=label)
    static_text.SetForegroundColour(color)
    static_text.SetFont(font(size, font_family))
    return static_text


# Holds the label, help, and input controls for a given parameter.
FormInput = namedtuple('FormInput', ['control', 'label', 'help'])
Parameter = namedtuple('Parameter', ['value', 'section', 'readableName',
                                     'helpTip', 'recommended_values', 'type'])


class Form(wx.Panel):
    """The Form class is a wx.Panel that creates controls/inputs for each
    parameter in the provided json file."""

    def __init__(self, parent,
                 json_file: str='bcipy/parameters/parameters.json',
                 load_file: str=None,
                 control_width: int=300, control_height: int=25, **kwargs):
        super(Form, self).__init__(parent, **kwargs)

        self.json_file = json_file
        self.load_file = json_file if not load_file else load_file
        self.control_size = (control_width, control_height)
        self.help_font_size = 12
        self.help_color = 'DARK SLATE GREY'
        with open(self.load_file) as f:
            data = f.read()

        config = json.loads(data)
        self.params = {k: Parameter(*v.values()) for k, v in config.items()}

        # TODO: group inputs by section

        self.createControls()
        self.bindEvents()
        self.doLayout()

    def createControls(self):
        """Create controls (inputs, labels, etc) for each item in the
        parameters file."""

        self.controls = {}
        for key, param in self.params.items():
            if param.type == 'bool':
                form_input = self.bool_input(param)
            elif 'path' in param.type:
                form_input = self.file_input(param)
            elif type(param.recommended_values) == list:
                form_input = self.selection_input(param)
            # TODO: NumCtrl for numeric input types.
            # from wx.lib.masked import NumCtrl
            # elif param['type'] in ['float', 'int']:
            else:
                form_input = self.text_input(param)

            self.add_input(self.controls, key, form_input)

        self.saveButton = wx.Button(self, label='Save')

    def add_input(self, controls: Dict[str, wx.Control], key: str,
                  form_input: FormInput) -> None:
        """Adds the controls for the given input to the controls structure."""
        if form_input.label:
            controls[f'{key}_label'] = form_input.label
        if form_input.help:
            controls[f'{key}_help'] = form_input.help
        controls[key] = form_input.control

        # TODO: consider adding an empty space after each input:
        # controls[f"{key}_empty"] = (0,0)

    def bool_input(self, param: Parameter) -> FormInput:
        """Creates a checkbox FormInput"""
        ctl = wx.CheckBox(self, label=param.readableName)
        ctl.SetValue(param.value == 'true')
        ctl.SetFont(font())
        return FormInput(ctl, label=None, help=None)

    def selection_input(self, param: Parameter) -> FormInput:
        """Creates a selection pulldown FormInput."""
        ctl = wx.ComboBox(self, -1, param.value,
                          size=self.control_size,
                          choices=param.recommended_values,
                          style=wx.CB_DROPDOWN)
        ctl.SetFont(font())
        label, help_tip = self.input_label(param)
        return FormInput(ctl, label, help_tip)

    def text_input(self, param: Parameter) -> FormInput:
        """Creates a text field FormInput."""
        ctl = wx.TextCtrl(self, size=self.control_size, value=param.value)
        ctl.SetFont(font())
        label, help_tip = self.input_label(param)
        return FormInput(ctl, label, help_tip)

    def file_input(self, param: Parameter) -> FormInput:
        """File Input.

        Creates a text field or selection pulldown FormInput with a button
        to browse for a file.
        """
        # Creates a combobox instead of text field if the parameter has recommended values
        if type(param.recommended_values) == list:
            ctl = wx.ComboBox(
                self,
                -1,
                param.value,
                size=self.control_size,
                choices=param.recommended_values,
                style=wx.CB_DROPDOWN)
        else:
            ctl = wx.TextCtrl(self, size=self.control_size, value=param.value)
        ctl.SetFont(font())
        btn = wx.Button(self, label='...', size=(self.control_size[1], self.control_size[1]))
        ctl_array = [ctl, btn]
        label, help_tip = self.input_label(param)
        return FormInput(ctl_array, label, help_tip)

    def input_label(self, param: Parameter) -> Tuple[wx.Control, wx.Control]:
        """Returns a label control and maybe a help Control if the help
        text is different than the label text."""
        label = static_text_control(self, label=param.readableName)
        help_tip = None
        if param.readableName != param.helpTip:
            help_tip = static_text_control(self, label=param.helpTip,
                                           size=self.help_font_size,
                                           color=self.help_color)
        return (label, help_tip)

    def bindEvents(self):
        """Bind event handlers to the controls to update the parameter values
        when the inputs change."""

        control_events = []
        for k, control in self.controls.items():
            # Only bind events for control inputs, not for label and help items
            if k in self.params:
                param = self.params[k]
                if param.type == 'bool':
                    control.Bind(wx.EVT_CHECKBOX, self.checkboxEventHandler(k))
                elif 'path' in param.type:
                    is_directory = False
                    if param.type == 'directorypath':
                        is_directory = True
                    control[0].Bind(wx.EVT_TEXT, self.textEventHandler(k))
                    control[1].Bind(wx.EVT_BUTTON, self.buttonEventHandler(k, is_directory))
                elif type(param.recommended_values) == list:
                    control.Bind(wx.EVT_COMBOBOX, self.selectEventHandler(k))
                    # allow user to input a value not in the select list.
                    control.Bind(wx.EVT_TEXT, self.textEventHandler(k))
                else:
                    control.Bind(wx.EVT_TEXT, self.textEventHandler(k))

        self.saveButton.Bind(wx.EVT_BUTTON, self.onSave)

    def doLayout(self):
        """Layout the controls using Sizers."""

        sizer = wx.BoxSizer(wx.VERTICAL)

        button_control_len = 1  # save button
        rowlen = len(self.controls.keys()) + button_control_len
        gridSizer = wx.FlexGridSizer(rows=rowlen, cols=1, vgap=10, hgap=10)

        noOptions = dict()

        # Add the controls to the grid:
        for control in self.controls.values():
            if isinstance(control, list):
                hbox = wx.BoxSizer(wx.HORIZONTAL)
                hbox.Add(control[0], flag=wx.RIGHT, border=5)
                hbox.Add(control[1], **noOptions)
                gridSizer.Add(hbox, **noOptions)
            else:
                gridSizer.Add(control, **noOptions)

        # Add the save button
        gridSizer.Add(self.saveButton, flag=wx.ALIGN_CENTER)

        sizer.Add(gridSizer, border=10, flag=wx.ALL | wx.ALIGN_CENTER)
        self.SetSizerAndFit(sizer)

    # Callback methods:
    def onSave(self, event: wx.EVT_BUTTON) -> None:
        log.debug('Saving parameter data')

        with open(self.json_file, 'w') as outfile:
            json.dump({k: v._asdict() for k, v in self.params.items()},
                      outfile, indent=JSON_INDENT)

        dialog = wx.MessageDialog(
            self, "Parameters Successfully Updated!", 'Info',
            wx.OK | wx.ICON_INFORMATION)
        dialog.ShowModal()
        dialog.Destroy()

    def textEventHandler(self, key: str) -> Callable[[wx.EVT_TEXT], None]:
        """Returns a handler function that updates the parameter for the
        provided key.
        """
        def handler(event: wx.EVT_TEXT):
            p = self.params[key]
            self.params[key] = p._replace(value=event.GetString())
            log.debug(f"{key}: {event.GetString()}")
        return handler

    def selectEventHandler(self, key: str) -> Callable[[wx.EVT_COMBOBOX], None]:
        """Returns a handler function that updates the parameter for the
        provided key.
        """
        def handler(event: wx.EVT_COMBOBOX):
            p = self.params[key]
            self.params[key] = p._replace(value=event.GetString())
            log.debug(f"{key}: {event.GetString()}")
        return handler

    def checkboxEventHandler(self, key: str) -> Callable[[wx.EVT_CHECKBOX], None]:
        """Returns a handler function that updates the parameter for the
        provided key.
        """
        def handler(event: wx.EVT_CHECKBOX):
            p = self.params[key]
            value = 'true' if event.IsChecked() else 'false'
            self.params[key] = p._replace(value=value)
            log.debug(f'{key}: {bool(event.IsChecked())}')
        return handler

    def buttonEventHandler(self, key: str, directory: bool) -> Callable[[wx.EVT_BUTTON], None]:
        """Returns a handler function that updates the parameter for the
        provided key.
        """
        def handler(event: wx.EVT_BUTTON):
            # Change dialog type depending on whether parameter requires a directory or file
            if directory:
                file_dialog = wx.DirDialog(self, 'Select a path', style=wx.FD_OPEN)
            else:
                file_dialog = wx.FileDialog(
                    self, 'Select a file', style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
            if file_dialog.ShowModal() == wx.ID_CANCEL:
                new_path = None
            else:
                new_path = str(file_dialog.GetPath())
                if directory:
                    # Add operating system separator character to end of directory path
                    new_path = new_path + sep
                log.debug(new_path)

            if new_path:
                for item_key, field in self.controls.items():
                    if item_key == key:
                        field[0].SetValue(new_path)
                parameter_key = self.params[key]
                self.params[key] = parameter_key._replace(value=new_path)
                log.debug(f'Selected path: {new_path}')

        return handler


class MainPanel(scrolled.ScrolledPanel):
    """Panel which contains the Form. Responsible for selecting the json data
     to load and handling scrolling."""

    def __init__(self, parent, title='BCI Parameters',
                 json_file='bcipy/parameters/parameters.json'):
        super(MainPanel, self).__init__(parent, -1)
        self.json_file = json_file
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox = vbox

        self.form = Form(self, json_file)
        vbox.Add(static_text_control(self, label=title, size=20),
                 0, wx.TOP | wx.ALIGN_CENTER_HORIZONTAL, border=5)
        vbox.AddSpacer(10)

        loading_box = wx.BoxSizer(wx.VERTICAL)
        loading_box.Add(static_text_control(self, label=f'Editing: {json_file}',
                                            size=14))
        self.loaded_from = static_text_control(
            self,
            label=f'Loaded from: {json_file}',
            size=14)
        loading_box.Add(self.loaded_from)
        loading_box.AddSpacer(10)

        self.loadButton = wx.Button(self, label='Load')
        self.loadButton.Bind(wx.EVT_BUTTON, self.onLoad)
        loading_box.Add(self.loadButton)

        # Used for displaying help messages to the user.
        self.flash_msg = static_text_control(self, label='', size=14,
                                             color='FOREST GREEN')
        loading_box.Add(self.flash_msg)

        vbox.Add(loading_box, 0, wx.ALL, border=10)
        vbox.AddSpacer(10)
        vbox.Add(self.form)
        self.SetSizer(vbox)
        self.SetupScrolling()

    def onLoad(self, event: wx.EVT_BUTTON) -> None:
        """Event handler to load the form data from a different json file."""
        log.debug('Loading parameters file')

        with wx.FileDialog(self, 'Open parameters file',
                           wildcard='JSON files (*.json)|*.json',
                           style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) as fd:
            if fd.ShowModal() == wx.ID_CANCEL:
                return     # the user changed their mind
            load_file = fd.GetPath()
            self.loaded_from.SetLabel(f'Loaded from: {load_file}')
            self.flash_msg.SetLabel('Click the Save button to persist these '
                                    'changes.')
            self.vbox.Hide(self.form)
        self.form = Form(self, json_file=self.json_file, load_file=load_file)
        self.vbox.Add(self.form)
        self.SetupScrolling()

    def OnChildFocus(self, event):
        event.Skip()


def main(title='BCI Parameters', size=(650, 550),
         json_file='bcipy/parameters/parameters.json'):
    """Set up the GUI components and start the main loop."""

    app = wx.App(0)
    frame = wx.Frame(None, wx.ID_ANY, size=size, title=title)
    fa = MainPanel(frame, title=title, json_file=json_file)
    frame.Show()
    app.MainLoop()


if __name__ == '__main__':
    main()



from tkinter import Tk
import numpy as np
import pandas as pd
import logging
from codecs import open as codecsopen
from json import load as jsonload
import pickle

from tkinter.filedialog import askopenfilename, askdirectory

log = logging.getLogger(__name__)


def _cast_parameters(parameters: dict) -> dict:
    """Cast to Value.

    Take in a parameters dictionary and coverts to a dictionary with type converted
        and extranous information removed.
    """
    new_parameters = {}
    for key, value in parameters.items():
        new_parameters[key] = cast_value(value)

    return new_parameters


def cast_value(value):
    """Cast Value.

    Takes in a value with a desired type and attempts to cast it to that type.
    """
    actual_value = str(value['value'])
    actual_type = value['type']

    try:
        if actual_type == 'int':
            new_value = int(actual_value)
        elif actual_type == 'float':
            new_value = float(actual_value)
        elif actual_type == 'bool':
            new_value = True if actual_value == 'true' else False
        elif actual_type == 'str' or 'path' in actual_type:
            new_value = str(actual_value)
        else:
            raise ValueError('Unrecognized value type')

    except Exception:
        raise ValueError(f'Could not cast {actual_value} to {actual_type}')

    return new_value


def load_json_parameters(path: str, value_cast: bool=False) -> dict:
    """Load JSON Parameters.

    Given a path to a json of parameters, convert to a dictionary and optionally
        cast the type.

    Expects the following format:
    "fake_data": {
        "value": "true",
        "section": "bci_config",
        "readableName": "Fake Data Sessions",
        "helpTip": "If true, fake data server used",
        "recommended_values": "",
        "type": "bool"
        }

    PARAMETERS
    ----------
    :param: path: string path to the parameters file.
    :param: value_case: True/False cast values to specified type.
    """
    # loads in json parameters and turns it into a dictionary
    try:
        with codecsopen(path, 'r', encoding='utf-8') as f:
            parameters = []
            try:
                parameters = jsonload(f)

                if value_cast:
                    parameters = _cast_parameters(parameters)
            except ValueError:
                raise ValueError(
                    "Parameters file is formatted incorrectly!")

        f.close()

    except IOError:
        raise IOError("Incorrect path to parameters given! Please try again.")

    return parameters


def load_experimental_data() -> str:
    # use python's internal gui to call file explorers and get the filename
    try:
        Tk().withdraw()  # we don't want a full GUI
        filename = askdirectory()  # show dialog box and return the path

    except Exception as error:
        raise error

    log.debug("Loaded Experimental Data From: %s" % filename)
    return filename


def load_signal_model(filename: str=None):
    # use python's internal gui to call file explorers and get the filename

    if not filename:
        try:
            Tk().withdraw()  # we don't want a full GUI
            filename = askopenfilename()  # show dialog box and return the path

        # except, raise error
        except Exception as error:
            raise error

    # load the signal_model with pickle
    signal_model = pickle.load(open(filename, 'rb'))

    return (signal_model, filename)


def load_csv_data(filename: str=None) -> str:
    if not filename:
        try:
            Tk().withdraw()  # we don't want a full GUI
            filename = askopenfilename()  # show dialog box and return the path

        except Exception as error:
            raise error

    # get the last part of the path to determine file type
    file_name = filename.split('/')[-1]

    if 'csv' not in file_name:
        raise Exception(
            'File type unrecognized. Please use a supported csv type')

    return filename


def read_data_csv(folder: str, dat_first_row: int=2, info_end_row: int=1) -> tuple:
    """ Reads the data (.csv) provided by the data acquisition
        Arg:
            folder(str): file location for the data
            dat_first_row(int): row with channel names
            info_end_row(int): final row related with daq. info
                where first row idx is '0'
        Return:
            raw_dat(ndarray[float]): C x N numpy array with samples
                where C is number of channels N is number of time samples
            channels(list[str]): channels used in DAQ
            stamp_time(ndarray[float]): time stamps for each sample
            type_amp(str): type of the device used for DAQ
            fs(int): sampling frequency
    """
    dat_file = pd.read_csv(folder, skiprows=dat_first_row)

    # Remove object columns (ex. BCI_Stimulus_Marker column)
    # TODO: might be better in use:
    # dat_file.select_dtypes(include=['float64'])
    numeric_dat_file = dat_file.select_dtypes(exclude=['object'])
    channels = list(numeric_dat_file.columns[1:]) # without timestamp column

    temp = pd.DataFrame.as_matrix(numeric_dat_file)
    stamp_time = temp[:, 0]
    raw_dat = temp[:, 1:temp.shape[1]].transpose()

    dat_file_2 = pd.read_csv(folder, nrows=info_end_row)
    type_amp = list(dat_file_2.axes[1])[1]
    fs = np.array(dat_file_2)[0][1]

    return raw_dat, stamp_time, channels, type_amp, fs


def load_txt_data() -> str:
    try:
        Tk().withdraw()  # we don't want a full GUI
        filename = askopenfilename()  # show dialog box and return the path
    except Exception as error:
        raise error

    file_name = filename.split('/')[-1]

    if 'txt' not in file_name:
        raise Exception(
            'File type unrecognized. Please use a supported text type')

    return filename



"""Defines utility classes and functions used by the acquisition module."""
import multiprocessing
import threading
import numpy as np


def mock_data(n_records: int, n_cols: int, low: int = -1000, high: int = 1000):
    """Generator for mock data that streams a finite number of
    records.

    Parameters:
    -----------
        n_records: number of records to generate
        n_cols: number of columns of each record
        low: optional; min value for each item; default is -1000
        high: optional; max value for each item; default is 1000

    Returns:
    --------
        A generator that yields arrays of random data
    """
    return (mock_record(n_cols, low, high) for _i in range(n_records))

# def mock_data_stream(n_records: int, length: int, low: int= -1000,
#     high: int = 1000):
#     """Generator for mock data that streams continuously.

#     Parameters:
#     -----------
#         n_records: number of records to generate
#         length: length of each record
#         low: optional; min value for each item; default is -1000
#         high: optional; max value for each item; default is 1000

#     Returns:
#     --------
#         A generator that yields arrays of random data
#     """
#     while True:
#         yield mock_record(length, low, high)


def mock_record(n_cols: int = 25, low: int = -1000, high: int = 1000):
    """Create an list of random data.

    Parameters:
    -----------
        n_cols - number of columns in the generated record.
        low - minimum value in each position
        high - maximum value in each position
    Returns:
    --------
        list of random values
    """
    return [np.random.uniform(low, high) for _cc in range(n_cols)]


class StoppableProcess(multiprocessing.Process):
    """Process class with a stop() method. The process itself has to check
    regularly for the running() condition.
    """

    def __init__(self, *args, **kwargs):
        super(StoppableProcess, self).__init__(*args, **kwargs)
        self._stopper = multiprocessing.Event()

    def stop(self):
        """Stop the process."""
        self._stopper.set()

    def running(self):
        """Test if the process is currently running."""
        return not self._stopper.is_set()

    def stopped(self):
        """Test if the process is stopped."""
        return self._stopper.is_set()


class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the running() condition.
    """

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stopper = threading.Event()

    def stop(self):
        """Stop the thread."""
        self._stopper.set()

    def running(self):
        """Test if the thread is currently running."""
        return not self._stopper.is_set()

    def stopped(self):
        """Test if the thread is currently stopped."""
        return self._stopper.is_set()



# pylint: disable=too-few-public-methods,no-self-use
"""Tests for datastream generator module"""
from builtins import next
import unittest
import pytest
from past.builtins import map, range
from mock import mock_open, patch
from bcipy.acquisition.datastream.generator import random_data, file_data
from bcipy.acquisition.util import mock_data


class CustomEncoder():
    """Encodes data by prefixing with the count."""

    def __init__(self):
        super(CustomEncoder, self).__init__()
        self.counter = 0

    def encode(self, data):
        """Encode the data."""
        self.counter += 1
        return (self.counter, data)


class TestGenerator(unittest.TestCase):
    """Tests for Generator"""

    def test_random_generator(self):
        """Test default parameters for random generator"""
        gen = random_data()
        data = [next(gen) for _ in range(100)]
        self.assertEqual(len(data), 100)

    def test_random_high_low_values(self):
        """Random generator should allow user to set value ranges."""
        channel_count = 10
        low = -100
        high = 100
        gen = random_data(low=-100, high=100,
                          channel_count=channel_count)
        data = [next(gen) for _ in range(100)]

        self.assertEqual(len(data), 100)

        for record in data:
            self.assertEqual(len(record), channel_count)
            for value in record:
                self.assertTrue(low <= value <= high)

    def test_random_with_custom_encoder(self):
        """Random generator should allow a custom encoder."""

        channel_count = 10
        gen = random_data(encoder=CustomEncoder(),
                          channel_count=channel_count)

        data = [next(gen) for _ in range(100)]

        self.assertEqual(len(data), 100)
        for _count, record in data:
            self.assertEqual(len(record), channel_count)

        self.assertEqual(data[0][0], 1)
        self.assertEqual(data[99][0], 100)

    def test_file_generator(self):
        """Should stream data from a file."""
        row_count = 100
        header = ['col1,col2,col3']
        data = list(mock_data(row_count, len(header)))
        rows = map(lambda x: ','.join(map(str, x)), data)
        test_data = '\n'.join(header + rows)

        with patch('bcipy.acquisition.datastream.generator.open',
                   mock_open(read_data=test_data), create=True):

            gen = file_data(filename='foo', header_row=1)
            generated_data = [next(gen) for _ in range(row_count)]

            for i, row in enumerate(generated_data):
                self.assertEqual(row, data[i])

    def test_file_generator_end(self):
        """Should throw an exception when all data has been consumed"""
        row_count = 10

        header = ['col1,col2,col3']
        data = list(mock_data(row_count, len(header)))
        rows = map(lambda x: ','.join(map(str, x)), data)
        test_data = '\n'.join(header + rows)

        with patch('bcipy.acquisition.datastream.generator.open',
                   mock_open(read_data=test_data), create=True):
            gen = file_data(filename='foo', header_row=1)
            # exhaust the generator
            for _ in range(row_count):
                next(gen)

            with pytest.raises(StopIteration):
                data.append(next(gen))

    def test_file_with_custom_encoder(self):
        """Should allow a custom encoder"""

        col_count = 3
        row_count = 100

        header = ['col1,col2,col3']
        data = [[float(cnum + rnum) for cnum in range(col_count)]
                for rnum in range(row_count)]
        rows = map(lambda x: ','.join(map(str, x)), data)
        test_data = '\n'.join(header + rows)

        with patch('bcipy.acquisition.datastream.generator.open',
                   mock_open(read_data=test_data), create=True):

            gen = file_data(
                filename='foo', header_row=1, encoder=CustomEncoder())
            generated_data = [next(gen) for _ in range(row_count)]

            for _count, record in generated_data:
                self.assertEqual(len(record), col_count)

            self.assertEqual(generated_data[0][0], 1)
            self.assertEqual(generated_data[99][0], 100)



"""Manages the creation and shutdown of the language model server"""
import logging
import os
import time
from typing import Dict, List
import docker
import requests
from bcipy.language_model.errors import DockerDownError, ConnectionErr, \
    StatusCodeError
log = logging.getLogger(__name__)
# pylint: disable=too-few-public-methods,too-many-arguments


class LmServerConfig():
    """Config for starting the language model server (docker container).
    Parameters:
    -----------
        image - docker image name
        host - host on which to start the server
        port - port to access server
        docker_port - port within the container
        volumes - mapping between local file and docker container file.
    """

    def __init__(self, image: str, host: str = "127.0.0.1",
                 port: int = 5000, docker_port: int = 5000,
                 volumes: Dict = None):
        self.image = image
        self.host = host
        self.port = port
        self.docker_port = docker_port
        self.volumes = volumes or {}
        self.command = "python server.py"

    def asdict(self):
        """Returns a dict representation of the class."""
        return docker_run_params(self)


def docker_run_params(config: LmServerConfig) -> Dict:
    """Dict used to parameterize docker-py's container run method."""
    return dict(image=config.image,
                command=config.command,
                ports={str(config.docker_port) +
                       '/tcp': (config.host, str(config.port))},
                volumes=docker_volumes(config.volumes),
                detach=True,
                remove=True)


def running_containers(client, server_config: LmServerConfig) -> List:
    """List the running Containers given the server_config"""
    return client.containers.list(filters={"ancestor": server_config.image})


def stop(server_config: LmServerConfig):
    """Stop the given docker image if it is currently running.
    """
    try:
        client = docker.from_env()
    except BaseException:
        raise DockerDownError  # docker ps for instance
    for con in running_containers(client, server_config):
        try:
            log.debug(f"Stopping existing container: {con.name}")
            con.stop()
            # Remove may throw an exception
            con.remove()
        except:
            pass


def start(server_config: LmServerConfig, max_wait: int = 16,
          start_complete_msg: str = "Running"):
    """Starts the language model server. This is currently implemented as a
    docker container that needs to be started, given the supplied config.
    Assumes that docker is installed and the correct server images have been
    loaded.

    Parameters:
    -----------
        server_config - configuration to startup the docker image
        max_wait - max seconds to wait during startup before returning.
        start_complete_msg - message that should be displayed in the (docker)
            logs when startup has completed.
    """
    try:
        client = docker.from_env()
    except BaseException:
        raise DockerDownError  # docker ps for instance

    # remove existing containers
    stop(server_config)

    run_params = server_config.asdict()
    t_start = time.time()
    try:
        container = client.containers.run(**run_params)
    except Exception:
        raise Exception("Error starting container. Try restarting docker")

    # wait for initialization
    log.debug("INITIALIZING LM SERVER.")
    log.debug(run_params)

    # wait for message in the logs or max_wait before continuing.
    sleep_interval = 0.5
    for _ in range(int(max_wait / sleep_interval)):
        if start_complete_msg in str(container.logs()):
            break
        time.sleep(sleep_interval)
    t_end = time.time()
    if start_complete_msg in str(container.logs()):
        log.debug(f"Container started in {t_end - t_start} seconds")
    else:
        log.debug("Wait time exceeded. Container may not be fully initialized.")

    # assert a new container was generated
    running_container_ids = [
        c.short_id for c in running_containers(client, server_config)]
    if not container.short_id in running_container_ids:
        raise Exception("Container did not correctly start.")


def post_json_request(server_config: LmServerConfig, path: str, data: Dict = None):
    """Posts a JSON request to the given url.
    Returns:
    --------
        json response if any or None.
    """
    data = data or {}
    host = server_config.host
    port = server_config.port
    url = f'http://{host}:{port}/{path}'
    try:
        response = requests.post(url, json=data)
    except requests.ConnectionError:
        raise ConnectionErr(host, port)
    if not response.status_code == requests.codes.ok:
        raise StatusCodeError(response.status_code)
    try:
        return response.json()
    except ValueError:
        # no JSON returned
        return None


def docker_volumes(volumes: Dict) -> Dict:
    """Converts {local_path: docker_path} dict to the format required by
    docker py. Checks validity of local path.
    """
    vols = {}
    for local_path, docker_path in volumes.items():
        assert os.path.exists(os.path.dirname(local_path)
                              ), "%r is not a valid path" % local_path
        vols[local_path] = {'bind': docker_path, 'mode': 'ro'}
    return vols


def main():
    """Starts a docker container for the given language model."""
    from bcipy.language_model.lm_modes import LmType
    import argparse

    lm_options = '; '.join(
        [f'{i+1} => {str(opt)}' for i, opt in enumerate(LmType)])
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--type', default=1, type=int,
                        help=f'Language model type. Options: ({lm_options})')
    args = parser.parse_args()

    config = LmType(args.type).model.DEFAULT_CONFIG
    start(config)


if __name__ == '__main__':
    main()



""" Wraps the Machine Learning Box which includes a classifier and a
dimensionality reduction method. It should form a pipeline to apply
different kinds of structures one after another.
"""


class Pipeline(object):
    """ Forms a pipeline using multiple dimensionality reductions and a
    final classifier. Observe that each function should include;
        - fit
        - transform
        - update
        - fit_transform
    Attr:
        pipeline(list[methods]): wrapper for pipeline structure
        line_el(list[ndarray[float]]): values at each input.
            [0] is the input and the rest follows
        """

    def __init__(self):
        self.pipeline = []
        self.line_el = []

    def add(self, method):
        """ Adds an object to the pipeline.
        Args:
            method(function): dim reduction or classification function """
        self.pipeline.append(method)

    def fit(self, x, y):
        """ Given the observation and the label, trains all elements in the
        pipeline. Observe that each element in the pipeline should have a
        'fit' function.
        Args:
            x(ndarray[float]): of desired shape
            y(ndarray[int]): of desired shape """
        self.line_el = [x]
        for i in range(len(self.pipeline) - 1):
            self.line_el.append(
                self.pipeline[i].fit_transform(self.line_el[i], y))

        self.pipeline[-1].fit(self.line_el[-1], y)

    def fit_transform(self, x, y):
        """ Applies fit transform on all functions
        Args:
            x(ndarray[float]): of desired shape
            y(ndarray[int]): of desired shape """

        self.line_el = [x]
        for i in range(len(self.pipeline) - 1):
            self.line_el.append(
                self.pipeline[i].fit_transform(self.line_el[i], y))

        arg = self.pipeline[-1].fit_transform(self.line_el[-1], y)
        return arg

    def transform(self, x):
        """ Applies transform on all functions. Prior to using transform on
        pipeline, it should be trained.
        Args:
             x(ndarray[float]): of desired shape """
        self.line_el = [x]
        for i in range(len(self.pipeline)):
            self.line_el.append(self.pipeline[i].transform(self.line_el[i]))

        return self.line_el[-1]



import numpy as np
from bcipy.signal.model.mach_learning.classifier.function_classifier \
    import RegularizedDiscriminantAnalysis
from bcipy.signal.model.mach_learning.dimensionality_reduction.function_dim_reduction \
    import ChannelWisePrincipalComponentAnalysis
from bcipy.signal.model.mach_learning.pipeline import Pipeline
from bcipy.signal.model.mach_learning.cross_validation import cross_validation, \
    cost_cross_validation_auc
from bcipy.signal.model.mach_learning.generative_mods.function_density_estimation \
    import KernelDensityEstimate
from sklearn import metrics
from scipy.stats import iqr

import logging

log = logging.getLogger(__name__)


def train_pca_rda_kde_model(x, y, k_folds=10):
    """ Trains the Cw-PCA RDA KDE model given the input data and labels with
        cross validation and returns the model
        Args:
            x(ndarray[float]): C x N x k data array
            y(ndarray[int]): N x 1 observation (class) array
                N is number of samples k is dimensionality of features
                C is number of channels
            k_folds(int): Number of cross validation folds
        Return:
            model(pipeline): trained likelihood model
            """

    # Pipeline is the model. It can be populated manually
    rda = RegularizedDiscriminantAnalysis()
    pca = ChannelWisePrincipalComponentAnalysis(var_tol=.1 ** 5,
                                                num_ch=x.shape[0])
    model = Pipeline()
    model.add(pca)
    model.add(rda)

    # Cross validate
    arg_cv = cross_validation(x, y, model=model, k_folds=k_folds)

    # Get the AUC before the regularization
    tmp, sc_cv, y_cv = cost_cross_validation_auc(model, 1, x, y, arg_cv,
                                                 k_folds=10, split='uniform')
    auc_init = -tmp
    # Start Cross validation
    lam = arg_cv[0]
    gam = arg_cv[1]
    log.debug('Optimized val [gam:{} \ lam:{}]'.format(lam, gam))
    model.pipeline[1].lam = lam
    model.pipeline[1].gam = gam
    tmp, sc_cv, y_cv = cost_cross_validation_auc(model, 1, x, y, arg_cv,
                                                 k_folds=10, split='uniform')
    auc_cv = -tmp

    # After finding cross validation scores do one more round to learn the final RDA model
    model.fit(x, y)

    # Insert the density estimates to the model and train using the cross validated
    # scores to avoid over fitting. Observe that these scores are not obtained using
    # the final model
    bandwidth = 1.06 * min(
        np.std(sc_cv), iqr(sc_cv) / 1.34) * np.power(x.shape[0], -0.2)
    model.add(KernelDensityEstimate(bandwidth=bandwidth))
    model.pipeline[-1].fit(sc_cv, y_cv)

    # Report AUC
    log.debug('AUC-i: {}, AUC-cv: {}'.format(auc_init, auc_cv))

    return model, auc_cv



# pylint: disable=fixme
"""Defines the driver for the Device for communicating with
LabStreamingLayer (LSL)."""
import logging

import pylsl
from bcipy.acquisition.protocols.device import Device

log = logging.getLogger(__name__)

LSL_TIMESTAMP = 'LSL_timestamp'


class Marker():
    """Data class which wraps a LSL marker; data pulled from a marker stream is
    a tuple where the first item is a list of channels and second item is the
    timestamp. Assumes that marker inlet only has a single channel."""

    def __init__(self, data=(None, None)):
        super(Marker, self).__init__()
        self.channels, self.timestamp = data

    @classmethod
    def empty(cls):
        """Creates an empty Marker."""
        return Marker()

    def __repr__(self):
        return f"<value: {self.trg}, timestamp: {self.timestamp}>"

    @property
    def is_empty(self):
        """Test to see if the current marker is empty."""
        return self.channels is None or self.timestamp is None

    @property
    def trg(self):
        """Get the trigger."""
        # pylint: disable=unsubscriptable-object
        return self.channels[0] if self.channels else None


def inlet_name(inlet) -> str:
    """Returns the name of a pylsl streamInlet."""
    name = '_'.join(inlet.info().name().split())
    return name.replace('-', '_')


class LslDevice(Device):
    """Driver for any device streaming data through the LabStreamingLayer lib.

    Parameters
    ----------
        connection_params : dict
            parameters used to connect with the server.
        channels: list, optional
            list of channel names
        fs: float, optional
            sample frequency in (Hz)
        include_lsl_timestamp: bool, optional
            if True, appends the LSL_timestamp to each sample.
        trg_channel_name: str, optional
            if present, the marker channel with this name is used as the TRG
            channel in the output and queried for offset; otherwise the last
            marker_channel added is used.
    """

    # pylint: disable=too-many-instance-attributes,too-many-arguments
    def __init__(self, connection_params, fs=None, channels=None,
                 include_lsl_timestamp=False,
                 trg_channel_name='BCI_Stimulus_Markers'):
        super(LslDevice, self).__init__(connection_params, fs, channels)

        self._appended_channels = []
        self.trg_channel_name = trg_channel_name

        if include_lsl_timestamp:
            self._appended_channels.append(LSL_TIMESTAMP)

        self._inlet = None
        self._marker_inlets = None
        # There can be 1 current marker for each marker channel.
        self.current_markers = {}

    @property
    def name(self):
        if 'stream_name' in self._connection_params:
            return self._connection_params['stream_name']
        if self._inlet and self._inlet.info().name():
            return self._inlet.info().name()
        return 'LSL'

    def connect(self):
        """Connect to the data source."""
        # Streams can be queried by name, type (xdf file format spec), and
        # other metadata.

        # NOTE: According to the documentation this is a blocking call that can
        # only be performed on the main thread in Linux systems. So far testing
        # seems fine when done in a separate multiprocessing.Process.
        eeg_streams = pylsl.resolve_stream('type', 'EEG')
        marker_streams = pylsl.resolve_stream('type', 'Markers')

        assert eeg_streams, "One or more EEG streams must be present"
        assert marker_streams, "One or more Marker streams must be present"
        self._inlet = pylsl.StreamInlet(eeg_streams[0])

        self._marker_inlets = [pylsl.StreamInlet(inlet)
                               for inlet in marker_streams]

        # initialize the current_markers for each marker stream.
        for inlet in self._marker_inlets:
            self.current_markers[inlet_name(inlet)] = Marker.empty()

    def acquisition_init(self):
        """Initialization step. Reads the channel and data rate information
        sent by the server and sets the appropriate instance variables.
        """
        assert self._inlet is not None, "Connect call is required."
        metadata = self._inlet.info()
        log.debug(metadata.as_xml())
        for marker_inlet in self._marker_inlets:
            log.debug("Streaming from marker inlet: %s",
                          inlet_name(marker_inlet))

        info_channels = self._read_channels(metadata)
        info_fs = metadata.nominal_srate()

        # If channels are not initially provided, set them from the metadata.
        # Otherwise, confirm that provided channels match metadata, or meta is
        # empty.
        if not self.channels:
            self.channels = info_channels
            assert self.channels, "Channels must be provided"
        else:
            if info_channels and self.channels != info_channels:
                raise Exception("Channels read from the device do not match "
                                "the provided parameters")
        assert len(self.channels) == (metadata.channel_count() +
                                      len(self._appended_channels) +
                                      len(self._marker_inlets)),\
            "Channel count error"

        if not self.fs:
            self.fs = info_fs
        elif self.fs != info_fs:
            raise Exception("Sample frequency read from device does not match "
                            "the provided parameter")

    def _trigger_inlet_index(self):
        """Index of the marker inlet that should be used for the TRG column.
        If the trg_channel_name parameter is provided, and there is a marker
        inlet with this name, that is used, otherwise the last marker inlet
        is used. Returns -1 if there are no marker inlets."""

        inlet_names = [inlet_name(inlet) for inlet in self._marker_inlets]

        if self.trg_channel_name in inlet_names:
            return inlet_names.index(self.trg_channel_name)
        return len(inlet_names) - 1

    def _read_channels(self, info):
        """Read channels from the stream metadata if provided and return them
        as a list. If channels were not specified, returns an empty list.

        Parameters
        ----------
            info : pylsl.XMLElement
        Returns
        -------
            list of str
        """
        channels = []
        if info.desc().child("channels").empty():
            return channels

        channel = info.desc().child("channels").child("channel")
        for _ in range(info.channel_count()):
            channel_name = channel.child_value("label")
            # If the data stream has a TRG channel, rename it so it doesn't
            # conflict with the marker channel.
            if channel_name == 'TRG' and self._marker_inlets:
                channel_name = "TRG_device_stream"
            channels.append(channel_name)
            channel = channel.next_sibling()

        for appended_channel in self._appended_channels:
            channels.append(appended_channel)

        trg_marker_index = self._trigger_inlet_index()
        for i, inlet in enumerate(self._marker_inlets):
            col = inlet_name(inlet)
            if i == trg_marker_index:
                col = 'TRG'
            channels.append(col)

        return channels

    def read_data(self):
        """Reads the next packet and returns the sensor data.

        Returns
        -------
            list with an item for each channel.
        """
        sample, timestamp = self._inlet.pull_sample()

        # Useful for debugging.
        if LSL_TIMESTAMP in self._appended_channels:
            sample.append(timestamp)

        for marker_inlet in self._marker_inlets:
            name = inlet_name(marker_inlet)
            marker = self.current_markers.get(name, Marker().empty())

            # Only attempt to retrieve a marker from the inlet if we have
            # merged the last one with a sample.
            if marker.is_empty:
                # A timeout of 0.0 only returns a sample if one is buffered for
                # immediate pickup. Without a timeout, this is a blocking call.
                marker_data = marker_inlet.pull_sample(timeout=0.0)
                marker = Marker(marker_data)
                self.current_markers[name] = marker

                if not marker.is_empty:
                    log.debug(
                        "Read marker %s from %s; current sample time: %s",
                        marker, name, timestamp)

            trg = "0"
            if not marker.is_empty and timestamp >= marker.timestamp:
                trg = marker.trg
                log.debug(("Appending %s marker %s to sample at time %s; ",
                               "time diff: %s"),
                              name, marker, timestamp,
                              timestamp - marker.timestamp)
                self.current_markers[name] = Marker.empty()  # clear current

            # Add marker field to sample
            sample.append(trg)

        return sample



import subprocess
import wx
from bcipy.gui.gui_main import BCIGui


# Make a custom GUI class with the events we want
class BCInterface(BCIGui):
    def bind_action(self, action: str, btn: wx.Button) -> None:
        if action == 'launch_mode':
            self.Bind(wx.EVT_BUTTON, self.launch_mode, btn)
        else:
            self.Bind(wx.EVT_BUTTON, self.on_clicked, btn)

    def launch_mode(self, event) -> None:
        mode_label = event.GetEventObject().GetLabel()
        if mode_label == 'RSVP':
            subprocess.Popen('python bcipy/gui/mode/RSVPKeyboard.py', shell=True)


app = wx.App(False)
gui = BCInterface(
    title="Brain Computer Interface",
    size=(700, 400), background_color='black')

# STATIC TEXT!
gui.add_static_text(
    text='Brain Computer Interface', position=(150, 0), size=25, color='white')

# BUTTONS!
gui.add_button(
    message="RSVP",
    position=(155, 200), size=(100, 100),
    color=wx.Colour(221, 37, 56),
    action='launch_mode')
gui.add_button(
    message="Matrix", position=(280, 200),
    size=(100, 100),
    color=wx.Colour(62, 161, 232),
    action='launch_mode')
gui.add_button(
    message="Shuffle", position=(405, 200),
    size=(100, 100),
    color=wx.Colour(117, 173, 48),
    action='launch_mode')

# Images
gui.add_image(
    path='bcipy/static/images/gui_images/ohsu.png', position=(5, 0), size=125)
gui.add_image(
    path='bcipy/static/images/gui_images/neu.png', position=(550, 0), size=125)


# Make the GUI Show now
gui.show_gui()
app.MainLoop()



from psychopy import visual
from bcipy.display.rsvp.display import RSVPDisplay
from bcipy.helpers.stimuli import resize_image
from bcipy.helpers.task import SPACE_CHAR


class IconToIconDisplay(RSVPDisplay):
    """ Icon to Icon or Icon to Word task display
        is_word determines whether this is an icon to word matching task"""

    def __init__(self, window, clock,
                 experiment_clock,
                 marker_writer,
                 info_text='Press Space Bar to Pause',
                 info_color='White',
                 info_pos=(0, -.9),
                 info_height=0.2,
                 info_font='Times',
                 task_color=['white'],
                 task_font='Times',
                 task_height=0.1,
                 stim_font='Times',
                 stim_pos=(-.8, .9),
                 stim_height=0.2,
                 stim_sequence=['a'] * 10,
                 stim_colors=['white'] * 10,
                 stim_timing=[1] * 10,
                 is_txt_stim=True,
                 trigger_type='image',
                 is_word=False):
        """ Initializes Icon Matching Task Objects """

        info_color = [info_color]
        info_font = [info_font]
        info_text = [info_text]
        info_pos = [info_pos]
        info_height = [info_height]

        if is_word:
            task_height *= 2

        tmp = visual.TextStim(win=window, font=task_font, text=' ')
        x_task_pos = tmp.boundingBox[0] / window.size[0] - 1

        self.task_pos = (x_task_pos, 1 - task_height)
        self.stim_height = stim_height
        self.stim_pos = stim_pos

        super(IconToIconDisplay, self).__init__(
            window, clock,
            experiment_clock,
            marker_writer,
            task_color=task_color,
            task_font=task_font,
            task_pos=self.task_pos,
            task_height=task_height,
            info_color=info_color,
            info_text=info_text,
            info_font=info_font,
            info_pos=info_pos,
            stim_font=stim_font,
            stim_pos=stim_pos,
            stim_height=stim_height,
            stim_sequence=stim_sequence,
            stim_colors=stim_colors,
            stim_timing=stim_timing,
            is_txt_stim=is_txt_stim,
            trigger_type=trigger_type)

        self.is_word = is_word

        if not is_word:
            self.rect = visual.Rect(
                win=window,
                width=task_height,
                height=task_height,
                lineColor='black',
                pos=(stim_pos),
                lineWidth=10,
                ori=0.0)
            self.rect_drawn_frames = 0

            self.task = visual.ImageStim(
                win=window, image=None, mask=None,
                units='', pos=self.task_pos,
                size=(task_height * 2, task_height * 2),
                ori=0.0)

            self.target_text = visual.TextStim(
                win=window,
                color='yellow',
                text='TARGET:',
                pos=(stim_pos[0] - 0.5, stim_pos[1]),
                height=task_height)

    def draw_static(self):
        if not self.is_word:
            """Draw static elements in a stimulus."""
            if(self.rect_drawn_frames < self.time_to_present):
                self.rect.draw()
                self.target_text.draw()
                self.rect_drawn_frames += 1

        super(IconToIconDisplay, self).draw_static()

    def update_task_state(
            self,
            image_path,
            task_height,
            rect_color,
            window_size,
            is_word):
        """ Updates task state of Icon to Icon/Word Matching Task by changing the
        image or text displayed at the top of the screen.
        Also updates rectangle size.
            Args:
                image_path(str): the path to the image to be displayed
                task_height(int): the height of the task image
                rect_color(str): the color of the rectangle
                window_size(tuple): The size of the window
                is_word(bool): word matching task
        """

        if is_word:
            # Display text at top of screen if we are matching icons to words
            txt = image_path if len(image_path) > 0 else ' '
            tmp2 = visual.TextStim(win=self.window, font=self.task.font, text=txt)
            x_task_pos = (tmp2.boundingBox[0] * 2.2) / self.window.size[0] - 1
            self.task_pos = (x_task_pos, self.task_pos[1])
            self.update_task(text=txt, color_list=['white'], pos=self.task_pos)
        else:
            # Otherwise, display an image at the top of the screen
            self.task.image = image_path

            image_width, image_height = resize_image(
                image_path,
                window_size,
                task_height)

            self.target_text.pos = (
                self.stim_pos[0] - image_width - 0.5,
                self.stim_pos[1])

            self.task.pos = (
                self.task_pos[0] + image_width * 2,
                self.task_pos[1] - image_width / 2)
            self.task.size = (image_width * 2, image_height * 2)

            self.rect_drawn_frames = 0
            self.rect.width = image_width / task_height * self.stim_height
            self.rect.height = image_height / task_height * self.stim_height
            self.rect.lineColor = rect_color



import numpy as np


def inference(x, targets, model, alphabet):
    """Evaluate the log likelihood ratios given the model and input.

    Then maps the distribution over the alphabet.
    Args:
        x(ndarray(float)): 3 dimensional np array first dimension is
            channels second dimension is trials and third dimension is
            time samples.
        targets(ndarray[str]): flashed symbols in order.
        model(pipeline): trained likelihood model.
        alphabet(list[str]): symbol set. Letters in the alphabet. All uppercase
    Return:
        lik_r(ndarray[float]): likelihood array.
    """

    # Evaluates the likelihood probabilities for p(e|l=1) and p(e|l=0)
    scores = np.exp(model.transform(x))
    # Evaluates the likelihood ratios
    scores = scores[:, 1] / (scores[:, 0] + np.power(.1, 10)) + np.power(.1, 10)
    # print("These values should be between 0 and 1: ", scores)

    # This maps the likelihood distribution over the alphabet
    #   If the letter in the alphabet does not exist in the target string,
    #       it takes 1
    lik_r = np.ones(len(alphabet))
    for idx in range(len(scores)):
        lik_r[alphabet.index(targets[idx])] *= scores[idx]

    return lik_r



from bcipy.feedback.feedback import Feedback
from psychopy import core
import sounddevice as sd


class AuditoryFeedback(Feedback):
    """Auditory Feedback."""
    def __init__(self, parameters, clock):

        # Register Feedback Type
        self.feedback_type = 'Auditory Feedback'

        super(AuditoryFeedback, self).__init__(self.feedback_type)

        # Parameters Dictionary
        self.parameters = parameters

        # Clock
        self.clock = clock

    def administer(self, sound, fs, assertion=None):
        timing = []

        if assertion:
            pass

        time = ['auditory_feedback', self.clock.getTime()]
        sd.play(sound, fs, blocking=True)
        core.wait(1)
        timing.append(time)

        return timing



import numpy as np


class RegularizedDiscriminantAnalysis:
    """ Regularized Discriminant Analysis for quadratic boundary in high
    dimensional spaces. Fits discriminant function,
        gi(x) = ln(pi(x)) - (1/2)(x-m)T E^(-1)(x-m) - ln(2pi|E|)
    uses -gi(x) as negative log probability for classification
    Ref:
        Friedman, Jerome H. "Regularized discriminant analysis."
        Journal of the American statistical association 84.405 (1989): 165-175
    Attr:
        lam(float): shrinkage param
        gam(float): threshold param (a.k.a. regularization param)
        class_i(list[int]): class labels
        mean_i(list[ndarray]): list of k x 1 dimensional mean vectors
        prior_i(list[float]): list of prior probabilities
        k (int): Number of features in one sample
        S (ndarray): k x k ndarray
            total data covariance matrix multiplied by number of data samples N
        N (int): Number of samples in x
        cov (ndarray): k x k ndarray, sample covariance of x
        S_i (list[ndarray]): k x k ndarray
            class covariance matrix multiplied by number of data samples N_i in class i
        N_i (list[int]): Number of samples in class i
        cov_i (list[ndarray]): list of k x k ndarray, sample covariance for class i
        log_det_cov(list[float]): list of negative det(cov_i)
        inv_reg_cov_i(list[ndarray]): inverse of regularized covariance matrix for class i

    """

    def __init__(self):  # TODO: Make it more modular
        self.lam = .9
        self.gam = .1

        self.class_i = None
        self.mean_i = None
        self.prior_i = None

        self.k = None

        self.S = None
        self.N = None

        self.S_i = None
        self.N_i = None
        self.cov_i = None

        self.log_det_reg_cov_i = None
        self.inv_reg_cov_i = None

    def fit(self, x, y, p=[]):
        """ Fits mean and covariance to the provided data
            and computes regularized covariances based on hyper parameters
            Args:
                x(ndarray[float]): N x k data array
                    N is number of samples k is dimensionality of features
                y(ndarray[int]): N x 1 observation (class) array
                p(ndarray[float]): c x 1 array with prior probabilities
                    c is number of classes in data
                """

        self.N, self.k = x.shape

        # Unique labels/classes
        self.class_i = np.unique(y)

        # Number of data samples in each class
        self.N_i = [np.sum(y == i)
                    for i in self.class_i]

        # MATLAB gets confused if np.where is not used. Insert this relation
        #  in order to make the ndarray readable from MATLAB side. There are
        #  two arrays, [0] for the correctness, choose it
        # Class means
        self.mean_i = [np.mean(x[np.where(y == i)[0]], axis=0)
                       for i in self.class_i]

        # Normalized x
        norm_vec = [x[np.where(y == self.class_i[i])[0]] - self.mean_i[i]
                    for i in range(len(self.class_i))]

        # Outer product of data matrix, Xi'Xi for each class
        self.S_i = [np.dot(np.transpose(norm_vec[i]), norm_vec[i])
                    for i in range(len(self.class_i))]

        # Sample covariances are calculated Si/Ni for each class
        self.cov_i = [self.S_i[i] / self.N_i[i]
                      for i in range(len(self.class_i))]

        # Sample covariance of total data
        self.S = np.zeros((self.k, self.k))

        for i in range(len(self.class_i)):
            self.S += self.S_i[i]

        # Set priors
        if len(p) == 0:
            prior = np.asarray([np.sum(y == self.class_i[i]) for i in
                                range(len(self.class_i))], dtype=float)
            self.prior_i = np.divide(prior, np.sum(prior))
        else:
            self.prior_i = p

        self.regularize(param=[self.gam, self.lam])

    def regularize(self, param):  # TODO: what if no param passed?
        """ Regularizes the covariance based on hyper parameters
            Args:
                param(list[gam(float),lam(float)]): List of regularization
                    parameters. Parameters should be a list instead of
                    individual elements for training purposes.
                 """

        self.gam = param[0]
        self.lam = param[1]

        # Shrinked class covariances
        shr_cov_i = [((1 - self.lam) * self.S_i[i] + self.lam * self.S) /
                     ((1 - self.lam) * self.N_i[i] + self.lam * self.N)
                     for i in range(len(self.class_i))]

        # Regularized class covariances
        reg_cov_i = [((1 - self.gam) * shr_cov_i[i] +
                      self.gam / self.k * np.trace(shr_cov_i[i]) *
                      np.eye(self.k)) for i in range(len(self.class_i))]

        self.inv_reg_cov_i, self.log_det_reg_cov_i = [], []

        # Use QR decomposition to find inverse of regularized covariance matrices
        # and their log of determinants
        for i in range(len(self.class_i)):
            q, r = np.linalg.qr(reg_cov_i[i])
            self.inv_reg_cov_i.append(np.linalg.solve(r, np.transpose(q)))
            # self.log_det_reg_cov_i.append(np.sum(np.log(np.abs(np.diag(r)))))

    def transform(self, x):

        val = self.get_prob(x)
        # as the val includes negative log likelihoods it outputs the
        # likelihood ratio for log(p(x|l=1)/p(x|l=0))
        if val.shape[1] == 2:
            val = val[:, 1] - val[:, 0]

        return val

    def get_prob(self, x):
        """ Gets -log likelihoods for each class
            Args:
                x(ndarray): N x k data array where
                    N is number of samples k is dimensionality of features
            Return:
                neg_log_l(ndarray): N x c negative log likelihood array
                    N is number of samples c is number of classes
                """

        neg_log_l = np.zeros([x.shape[0], len(self.class_i)])
        for s in range(x.shape[0]):
            for i in range(len(self.class_i)):
                zero_mean = x[s] - self.mean_i[i]

                # Every constant at the end of score calculation is omitted.
                # This is why we omit log det of class regularized covariances.
                evidence = np.dot(zero_mean,
                                  np.dot(self.inv_reg_cov_i[i], zero_mean))

                neg_log_l[s][i] = -.5 * evidence + np.log(self.prior_i[i])

        return neg_log_l

    def fit_transform(self, x, y, p=[]):
        """ Fits the model to provided (x,y) = (data,obs) couples and
        returns the negative log likelihoods.
            Args:
                x(ndarray[float]): N x k data array
                    N is number of samples k is dimensionality of features
                y(ndarray[int]): N x 1 observation (class) array
                p(ndarray[float]): c x 1 array with prior probabilities
                    c is number  of classes in data
            Return:
                val(ndarray[float]): N x c negative log likelihood array
                """

        self.fit(x, y, p)

        return self.transform(x)



"""Defines the driver for the Device for communicating with the DSI headset."""
import logging
import socket

from bcipy.acquisition.protocols.dsi import dsi
from bcipy.acquisition.protocols import util
from bcipy.acquisition.protocols.device import Device

log = logging.getLogger(__name__)


class DsiDevice(Device):
    """Driver for the DSI device.

    Parameters
    ----------
        connection_params : dict
            parameters used to connect with the server. keys: [host, port]
        channels: list, optional
            list of channel names
        fs: float, optional
            sample frequency in (Hz)
    """

    def __init__(self, connection_params, fs=dsi.DEFAULT_FS, channels=None):
        """Init DsiDevice."""
        channels = channels if channels is not None else dsi.DEFAULT_CHANNELS
        super(DsiDevice, self).__init__(connection_params, fs, channels)
        assert 'host' in connection_params, "Please specify host to Device!"
        assert 'port' in connection_params, "Please specify port to Device!"
        self._channels_provided = len(channels) > 0
        self._socket = None
        self.channels = channels

    @property
    def name(self):
        """DSI Name."""
        return 'DSI'

    def connect(self):
        """Connect to the data source."""
        params = self._connection_params
        address = (params['host'], params['port'])
        self._socket = socket.create_connection(address)
        self._socket.settimeout(None)

    def disconnect(self):
        self._socket.close()

    def _read_packet(self):
        """Read a single packet from the data source.

        Returns
        -------
            dict-like object
        """

        assert self._socket is not None, \
            "Socket isn't started, cannot read DSI packet!"

        # Reads the header to get the payload length, then reads the payload.
        header_buf = util.receive(self._socket, dsi.HEADER_LEN)
        header = dsi.header.parse(header_buf)
        payload_buf = util.receive(self._socket, header.payload_length)
        return dsi.packet.parse(header_buf + payload_buf)

    def acquisition_init(self):
        """Initialization step.

        Reads the channel and data rate information
        sent by the server and sets the appropriate instance variables.
        """

        response = self._read_packet()
        # Read packets until we encounter the headers we care about.
        while response.type != 'EVENT' or response.event_code != 'SENSOR_MAP':
            if response.type == 'EEG_DATA':
                raise Exception('EEG Data was encountered; expected '
                                'initialization headers.')
            log.debug(response.type)

            # Here we get information from the device about version etc.
            #  If interested, print the response type and message!
            if response.type == 'EVENT':
                pass
            response = self._read_packet()

        channels = response.message.split(',')
        log.debug("Channels: %s", ','.join(channels))
        if self._channels_provided and len(channels) != len(self.channels):
            raise Exception("Channels read from DSI device do not match "
                            "the provided parameters")
        else:
            self.channels = dsi.DEFAULT_CHANNELS
        response = self._read_packet()

        if response.type != 'EVENT' or response.event_code != 'DATA_RATE':
            raise Exception("Unexpected packet; expected DATA RATE Event")

        sample_freq = int(response.message.split(',')[1])
        log.debug("Sample frequency: %s", str(sample_freq))

        if sample_freq != self.fs:
            raise Exception("Sample frequency read from DSI device does not "
                            "match the provided parameter")

        # Read once more for data start
        response = self._read_packet()

    def read_data(self):
        """Read Data.

        Reads he next packet from DSI device and returns the sensor data.

        Returns
        -------
            list with an item for each channel.
        """
        sensor_data = False

        while not sensor_data:

            response = self._read_packet()

            if hasattr(response, 'sensor_data'):
                return list(response.sensor_data)



"""Code for constructing and executing Tasks"""
from bcipy.tasks.rsvp.calibration.alert_tone_calibration import RSVPAlertToneCalibrationTask
from bcipy.tasks.rsvp.calibration.inter_sequence_feedback_calibration import (
    RSVPInterSequenceFeedbackCalibration
)
from bcipy.tasks.rsvp.calibration.calibration import RSVPCalibrationTask
from bcipy.tasks.rsvp.copy_phrase import RSVPCopyPhraseTask
from bcipy.tasks.rsvp.icon_to_icon import RSVPIconToIconTask

from bcipy.tasks.task import Task
from bcipy.tasks.exceptions import TaskRegistryException
from bcipy.tasks.task_registry import ExperimentType


def make_task(display_window, daq, exp_type, parameters, file_save,
              signal_model=None, language_model=None, fake=True,
              auc_filename=None) -> Task:
    """Creates a Task based on the provided parameters.

    Parameters:
    -----------
        display_window: pyschopy Window
        daq: DataAcquisitionClient
        exp_type: ExperimentType
        parameters: dict
        file_save: str - path to file in which to save data
        signal_model
        language_model - language model
        fake: boolean - true if eeg stream is randomly generated
        auc_filename: str
    Returns:
    --------
        Task instance
    """

    # NORMAL RSVP MODES
    if exp_type is ExperimentType.RSVP_CALIBRATION:
        return RSVPCalibrationTask(
            display_window, daq, parameters, file_save)

    if exp_type is ExperimentType.RSVP_COPY_PHRASE:
        return RSVPCopyPhraseTask(
            display_window, daq, parameters, file_save, signal_model,
            language_model, fake=fake)

    # ICON TASKS
    if exp_type is ExperimentType.RSVP_ICON_TO_ICON:
        return RSVPIconToIconTask(display_window, daq,
                                  parameters, file_save, signal_model,
                                  language_model, fake, False, auc_filename)

    if exp_type is ExperimentType.RSVP_ICON_TO_WORD:
        # pylint: disable=fixme
        # TODO: consider a new class for this scenario.
        return RSVPIconToIconTask(display_window, daq,
                                  parameters, file_save, signal_model,
                                  language_model, fake, True, auc_filename)

    # CALIBRATION FEEDBACK TASKS
    if exp_type is ExperimentType.RSVP_ALERT_TONE_CALIBRATION:
        return RSVPAlertToneCalibrationTask(
            display_window, daq, parameters, file_save)

    if exp_type is ExperimentType.RSVP_INTER_SEQUENCE_FEEDBACK_CALIBRATION:
        return RSVPInterSequenceFeedbackCalibration(
            display_window, daq, parameters, file_save)

    raise TaskRegistryException('The provided experiment type is not registered.')


def start_task(display_window, daq, exp_type, parameters, file_save,
               signal_model=None, language_model=None, fake=True, auc_filename=None):
    """Creates a Task and starts execution."""
    task = make_task(display_window, daq, exp_type, parameters, file_save,
                     signal_model, language_model, fake, auc_filename)
    task.execute()



from bcipy.helpers.load import load_txt_data
from bcipy.helpers.stimuli import resize_image, play_sound
import csv
from typing import TextIO, List, Tuple

from psychopy import visual, core

NONE_VALUE = '0'
SOUND_TYPE = 'sound'
IMAGE_TYPE = 'image'


class TriggerCallback:
    timing = None
    first_time = True

    def callback(self, clock: core.Clock, stimuli: str) -> None:
        if self.first_time:
            self.timing = [stimuli, clock.getTime()]
            self.first_time = False

    def reset(self):
        self.timing = None
        self.first_time = True


def _calibration_trigger(experiment_clock: core.Clock,
                         trigger_type: str='sound',
                         trigger_name: str='calibration_trigger',
                         display=None,
                         on_trigger=None) -> List[tuple]:
    """Calibration Trigger.

        Outputs triggers for the purpose of calibrating data and stimuli.
        This is an ongoing difficulty between OS, DAQ devices and stimuli type. This
        code aims to operationalize the approach to finding the correct DAQ samples in
        relation to our trigger code.

    PARAMETERS
    ---------
        experiment_clock(clock): clock with getTime() method, which is used in the code
            to report timing of stimuli
        trigger_type(string): type of trigger that is desired (sound, image, etc)
        display(DisplayWindow): a window that can display stimuli. Currently, a Psychopy window.
        on_trigger(function): optional callback; if present gets called
                 when the calibration trigger is fired; accepts a single
                 parameter for the timing information.
        Return:
            timing(array): timing values for the calibration triggers to be written to trigger file or
                    used to calculate offsets.
    """
    trigger_callback = TriggerCallback()

    # If sound trigger is selected, output calibration tones
    if trigger_type == SOUND_TYPE:
        import sounddevice as sd
        import soundfile as sf

        play_sound(
            sound_file_path='bcipy/static/sounds/1k_800mV_20ms_stereo.wav',
            dtype='float32',
            track_timing=True,
            sound_callback=trigger_callback,
            sound_load_buffer_time=0.5,
            experiment_clock=experiment_clock,
            trigger_name='calibration_trigger')

    elif trigger_type == IMAGE_TYPE:
        if display:
            calibration_box = visual.ImageStim(
                win=display,
                image='bcipy/static/images/testing_images/white.png',
                pos=(-.5, -.5),
                mask=None,
                ori=0.0)
            calibration_box.size = resize_image(
                'bcipy/static/images/testing_images/white.png',
                display.size, 0.75)

            display.callOnFlip(trigger_callback.callback, experiment_clock, trigger_name)
            if on_trigger is not None:
                display.callOnFlip(on_trigger, trigger_name)

            presentation_time = int(1 * display.getActualFrameRate())
            for num_frames in range(presentation_time):
                calibration_box.draw()
                display.flip()

        else:
            raise Exception(
                'Display object required for calibration with images!')

    else:
        raise Exception('Trigger type not implemented for Calibration yet!')

    return trigger_callback.timing


def _write_triggers_from_sequence_calibration(
        array: list,
        trigger_file: TextIO,
        offset: bool=False):
    """Write triggers from calibration.

    Helper Function to write trigger data to provided trigger_file. It assigns
        target letter based on the first presented letter in sequence, then
        assigns target/nontarget label to following letters.

    It writes in the following order:
        (I) presented letter, (II) targetness, (III) timestamp
    """

    x = 0

    if offset:
        # extract the letter and timing from the array
        (letter, time) = array
        targetness = 'offset_correction'
        trigger_file.write('%s %s %s' % (letter, targetness, time) + "\n")

    else:
        for i in array:

            # extract the letter and timing from the array
            (letter, time) = i

            # determine what the trigger are
            if letter == 'calibration_trigger':
                targetness = 'calib'
                target_letter = letter
            else:
                if x == 0:
                    targetness = 'first_pres_target'
                    target_letter = letter
                elif x == 1:
                    targetness = 'fixation'
                elif x > 1 and target_letter == letter:
                    targetness = 'target'
                else:
                    targetness = 'nontarget'

                x += 1

            # write to the trigger_file
            trigger_file.write('%s %s %s' % (letter, targetness, time) + "\n")

    return trigger_file


def _write_triggers_from_sequence_copy_phrase(
        array,
        trigger_file,
        copy_text,
        typed_text,
        offset=None):
    """
    Write triggers from copy phrase.

    Helper Function to write trigger data to provided trigger_file. It assigns
        target letter based on matching the next needed letter in typed text
        then assigns target/nontarget label to following letters.

    It writes in the following order:
        (I) presented letter, (II) targetness, (III) timestamp
    """

    if offset:
        # extract the letter and timing from the array
        (letter, time) = array
        targetness = 'offset_correction'
        trigger_file.write('%s %s %s' % (letter, targetness, time) + "\n")

    else:
        # get relevant spelling info to determine what was and should be typed
        spelling_length = len(typed_text)
        last_typed = typed_text[-1] if typed_text else None
        correct_letter = copy_text[spelling_length - 1]

        # because there is the impassibility of incorrect letter and correction,
        # we check here what is appropriate as a correct response
        if last_typed == correct_letter:
            correct_letter = copy_text[spelling_length]
        else:
            correct_letter = '<'

        x = 0

        for i in array:

            # extract the letter and timing from the array
            (letter, time) = i

            # determine what the triggers are:
            #       assumes there is no target letter presentation.
            if x == 0:
                targetness = 'fixation'
            elif x > 1 and correct_letter == letter:
                targetness = 'target'
            else:
                targetness = 'nontarget'

            # write to the trigger_file
            trigger_file.write('%s %s %s' % (letter, targetness, time) + "\n")

            x += 1

    return trigger_file


def _write_triggers_from_sequence_free_spell(array, trigger_file):
    """
    Write triggers from free spell.

    Helper Function to write trigger data to provided trigger_file.

    It writes in the following order:
        (I) presented letter, (II) timestamp
    """

    for i in array:

        # extract the letter and timing from the array
        (letter, time) = i

        # write to trigger_file
        trigger_file.write('%s %s' % (letter, time) + "\n")

    return trigger_file


def write_triggers_from_sequence_icon_to_icon(
        sequence_timing: List[Tuple], trigger_file: TextIO, target: str,
        target_displayed: bool, offset=None):
    """
    Write triggers from icon to icon task.
    It writes in the following order:
        (I) presented letter, (II) targetness, (III) timestamp

    Parameters:
    ----------
        sequence_timing - list of (icon, time) output from rsvp after
            displaying a sequence.
        trigger_file - open file in which to write.
        target - target for the current sequence
        target_displayed - whether or not the target was presented during the
            sequence.
    """
    if offset:
        (letter, time) = sequence_timing
        targetness = 'offset_correction'
        trigger_file.write('%s %s %s' % (letter, targetness, time) + "\n")
        return

    icons, _times = zip(*sequence_timing)
    calib_presented = 'calibration_trigger' in icons
    calib_index = 0 if calib_presented else -1

    if calib_presented:
        target_pres_index = 1
        fixation_index = 2
    elif target_displayed:
        target_pres_index = 0
        fixation_index = 1
    else:
        target_pres_index = -1
        fixation_index = 0

    for i, (icon, presentation_time) in enumerate(sequence_timing):
        targetness = 'nontarget'
        if i == calib_index:
            targetness = 'calib'
        elif i == target_pres_index:
            targetness = 'first_pres_target'
        elif i == fixation_index:
            targetness = 'fixation'
        elif icon == target:
            targetness = 'target'
        else:
            targetness = 'nontarget'
        trigger_file.write('%s %s %s' % (icon, targetness, presentation_time) +
                           "\n")


def trigger_decoder(mode: str, trigger_path: str=None) -> tuple:
    """Trigger Decoder.

    Given a mode of operation (calibration, copy phrase, etc) and
        a path to the trigger location (*.txt file), this function
        will split into symbols (A, ..., Z), timing info (32.222), and
        targetness (target, nontarget). It will also extract any saved
        offset information and pass that back.

    PARAMETERS
    ----------
    :param: mode: mode of bci operation. Note the mode changes how triggers
        are saved.
    :param: trigger_path: [Optional] path to triggers.txt file
    :return: tuple: symbol_info, trial_target_info, timing_info, offset.
    """

    # Load triggers.txt
    if not trigger_path:
        trigger_path = load_txt_data()

    # Get every line of trigger.txt
    with open(trigger_path, 'r+') as text_file:
        # most trigger files has three columns:
        #   SYMBOL, TARGETNESS_INFO[OPTIONAL], TIMING
        trigger_txt = [line.split() for line in text_file]

    # extract stimuli from the text
    stimuli_triggers = [line for line in trigger_txt
                        if line[1] == 'target' or
                        line[1] == 'nontarget']

    # from the stimuli array, pull our the symbol information
    symbol_info = list(map(lambda x: x[0], stimuli_triggers))

    # If operating mode is free spell, it only has 2 columns
    #   otherwise, it has 3
    if mode != 'free_spell':
        trial_target_info = list(map(lambda x: x[1], stimuli_triggers))
        timing_info = list(map(lambda x: eval(x[2]), stimuli_triggers))
    else:
        trial_target_info = None
        timing_info = list(map(lambda x: eval(x[1]), stimuli_triggers))

    # Get any offset or calibration triggers
    offset_array = [line[2] for line in trigger_txt
                    if line[0] == 'offset']
    calib_trigger_array = [line[2] for line in trigger_txt
                           if line[0] == 'calibration_trigger']

    # If present, calculate the offset between the DAQ and Triggers from display
    if len(offset_array) == 1 and len(calib_trigger_array) == 1:

        # Extract the offset and calibration trigger time
        offset_time = float(offset_array[0])
        calib_trigger_time = float(calib_trigger_array[0])

        # Calculate the offset (ASSUMES DAQ STARTED FIRST!)
        offset = offset_time - calib_trigger_time

    # Otherwise, assume no observed offset
    else:
        offset = 0

    return symbol_info, trial_target_info, timing_info, offset


class Labeller(object):
    """Labels the targetness for a trigger value in a raw_data file."""

    def __init__(self):
        super(Labeller, self).__init__()

    def label(self, trigger):
        raise NotImplementedError('Subclass must define the label method')


class LslCalibrationLabeller(Labeller):
    """Calculates targetness for calibration data. Uses a state machine to
    determine how to label triggers.

    Parameters:
    -----------
        seq_len: stim_length parameter value for the experiment; used to calculate
            targetness for first_pres_target.
    """

    def __init__(self, seq_len: int):
        super(LslCalibrationLabeller, self).__init__()
        self.seq_len = seq_len
        self.prev = None
        self.current_target = None
        self.seq_position = 0

    def label(self, trigger):
        """Calculates the targetness for the given trigger, accounting for the
        previous triggers/states encountered."""
        state = ''
        if self.prev is None:
            # First trigger is always calibration.
            state = 'calib'
        elif self.prev == 'calib':
            self.current_target = trigger
            state = 'first_pres_target'
        elif self.prev == 'first_pres_target':
            # reset the sequence when fixation '+' is encountered.
            self.seq_pos = 0
            state = 'fixation'
        else:
            self.seq_pos += 1
            if self.seq_pos > self.seq_len:
                self.current_target = trigger
                state = 'first_pres_target'
            elif trigger == self.current_target:
                state = 'target'
            else:
                state = 'nontarget'
        self.prev = state
        return state


class LslCopyPhraseLabeller(Labeller):
    """Sequentially calculates targetness for copy phrase triggers."""

    def __init__(self, copy_text: str, typed_text: str):
        super(LslCopyPhraseLabeller, self).__init__()
        self.copy_text = copy_text
        self.typed_text = typed_text
        self.prev = None

        self.pos = 0
        self.typing_pos = -1  # sequence length should be >= typed text.

        self.current_target = None

    def label(self, trigger):
        """Calculates the targetness for the given trigger, accounting for the
        previous triggers/states encountered."""
        state = ''
        if self.prev is None:
            state = 'calib'
        elif trigger == '+':
            self.typing_pos += 1
            if not self.current_target:
                # set target to first letter in the copy phrase.
                self.current_target = self.copy_text[self.pos]
            else:
                last_typed = self.typed_text[self.typing_pos - 1]
                if last_typed == self.current_target:
                    # increment if the user typed the target correctly
                    if last_typed != '<':
                        self.pos += 1
                    self.current_target = self.copy_text[self.pos]
                else:
                    # Error correction.
                    self.current_target = '<'

            state = 'fixation'
        else:
            if trigger == self.current_target:
                state = 'target'
            else:
                state = 'nontarget'
        self.prev = state
        return state


def _extract_triggers(csvfile: TextIO,
                      trg_field,
                      labeller: Labeller) -> List[Tuple[str, str, str]]:
    """Extracts trigger data from an experiment output csv file.
    Parameters:
    -----------
        csvfile: open csv file containing data.
        trg_field: optional; name of the data column with the trigger data;
                   defaults to 'TRG'
        labeller: Labeller used to calculate the targetness value for a
            given trigger.
    Returns:
    --------
        list of tuples of (trigger, targetness, timestamp)
    """
    data = []

    # Skip metadata rows
    _daq_type = next(csvfile)
    _sample_rate = next(csvfile)

    reader = csv.DictReader(csvfile)

    for row in reader:
        trg = row[trg_field]
        if trg != NONE_VALUE:
            if 'calibration' in trg:
                trg = 'calibration_trigger'
            targetness = labeller.label(trg)
            data.append((trg, targetness, row['timestamp']))

    return data


def write_trigger_file_from_lsl_calibration(csvfile: TextIO,
                                            trigger_file: TextIO,
                                            seq_len: int, trg_field: str='TRG'):
    """Creates a triggers.txt file from TRG data recorded in the raw_data
    output from a calibration."""
    extracted = extract_from_calibration(csvfile, seq_len, trg_field)
    _write_trigger_file_from_extraction(trigger_file, extracted)


def write_trigger_file_from_lsl_copy_phrase(csvfile: TextIO,
                                            trigger_file: TextIO,
                                            copy_text: str, typed_text: str,
                                            trg_field: str='TRG'):
    """Creates a triggers.txt file from TRG data recorded in the raw_data
    output from a copy phrase."""
    extracted = extract_from_copy_phrase(csvfile, copy_text, typed_text,
                                         trg_field)
    _write_trigger_file_from_extraction(trigger_file, extracted)


def _write_trigger_file_from_extraction(trigger_file: TextIO,
                                        extraction: List[Tuple[str, str, str]]):
    """Writes triggers that have been extracted from a raw_data file to a
    file."""
    for trigger, targetness, timestamp in extraction:
        trigger_file.write(f"{trigger} {targetness} {timestamp}\n")

    # TODO: is this assumption correct?
    trigger_file.write("offset offset_correction 0.0")


def extract_from_calibration(csvfile: TextIO,
                             seq_len: int,
                             trg_field: str='TRG') -> List[Tuple[str, str, str]]:
    """Extracts trigger data from a calibration output csv file.
    Parameters:
    -----------
        csvfile: open csv file containing data.
        seq_len: stim_length parameter value for the experiment; used to calculate
                 targetness for first_pres_target.
        trg_field: optional; name of the data column with the trigger data;
                   defaults to 'TRG'
    Returns:
    --------
        list of tuples of (trigger, targetness, timestamp), where timestamp is
        the timestamp recorded in the file.
    """

    return _extract_triggers(csvfile, trg_field,
                             labeller=LslCalibrationLabeller(seq_len))


def extract_from_copy_phrase(csvfile: TextIO,
                             copy_text: str,
                             typed_text: str,
                             trg_field: str='TRG') -> List[Tuple[str, str, str]]:
    """Extracts trigger data from a copy phrase output csv file.
    Parameters:
    -----------
        csvfile: open csv file containing data.
        copy_text: phrase to copy
        typed_text: participant typed response
        trg_field: optional; name of the data column with the trigger data;
                   defaults to 'TRG'
    Returns:
    --------
        list of tuples of (trigger, targetness, timestamp), where timestamp is
        the timestamp recorded in the file.
    """
    labeller = LslCopyPhraseLabeller(copy_text, typed_text)
    return _extract_triggers(csvfile, trg_field, labeller=labeller)



# pylint: disable=fixme,invalid-name
"""
Binary protocol for DSI; ported to construct 2.8 from the original
rsvpkeyboard app.
"""

from construct import (Array, Bytes, Const, Embedded, Enum, Float32b, Int8ub,
                       Int16ub, Int32ub, Optional, PascalString, Struct,
                       Switch, this)
DEFAULT_FS = 300
DEFAULT_CHANNELS = ['P3', 'C3', 'F3', 'Fz', 'F4', 'C4', 'P4', 'Cz',
                    'CM', 'A1', 'Fp1', 'Fp2', 'T3', 'T5', 'O1', 'O2',
                    'X3', 'X2', 'F7', 'F8', 'X1',
                    'A2', 'T6', 'T4', 'TRG']
header = Struct(
    'magic' / Const(b'@ABCD'),                           # bytes 0-5
    'type' / Enum(Int8ub, NULL=0, EEG_DATA=1, EVENT=5),  # byte 5
    'payload_length' / Int16ub,                          # bytes 6-7
    'number' / Int32ub,                                  # bytes 8-11
)

HEADER_LEN = 12

# TODO: Resolve the 'unknown encoding' error for the message when we import
# unicode_literals.
event = Struct(
    'event_code' / Enum(Int32ub,  # bytes 12-15
                        VERSION=1,
                        DATA_START=2,
                        DATA_STOP=3,
                        SENSOR_MAP=9,
                        DATA_RATE=10
                        ),
    'sending_node' / Int32ub,  # bytes 16-19

    # Message data is optional
    # Previous implementation used If(this._.payload_length > 8, ...), but this
    # method does not seem to work in v2.8.

    # message_length: bytes 20-23, message: bytes 24+
    'message' / Optional(PascalString(lengthfield='message_length' / Int32ub,
                                      encoding='ascii'))
)

EEG_data = Struct(
    'timestamp' / Float32b,   # bytes 12-15
    'data_counter' / Int8ub,  # byte 16; Unused, just 0 currently
    'ADC_status' / Bytes(6),  # bytes 17-22
    # bytes 23-26, 27-30, etc.
    'sensor_data' / Array((this._.payload_length - 11) // 4, Float32b)
)

null = Struct('none' / Array(111, Int8ub))

packet = Struct(
    Embedded(header),
    'payload' / Embedded(Switch(this.type,
                                {
                                    'NULL': null,
                                    'EVENT': event,
                                    'EEG_DATA': EEG_data
                                }))
)



import glob
import unittest
from os import path
import sounddevice as sd
import soundfile as sf
from mockito import any, mock, unstub, verify, when
from psychopy import core

from bcipy.helpers.stimuli import play_sound, soundfiles,\
 random_rsvp_calibration_seq_gen, generate_icon_match_images, best_selection,\
 best_case_rsvp_seq_gen

MOCK_FS = 44100


class TestStimuliGeneration(unittest.TestCase):
    """This is Test Case for Stimuli Generated via BciPy."""

    def tearDown(self):
        unstub()

    def test_play_sound_returns_timing(self):
        # fake sound file path
        sound_file_path = 'test_sound_file_path'

        # mock the other library interactions
        when(sf).read(
            sound_file_path, dtype='float32').thenReturn(('data', MOCK_FS))
        when(sd).play(any(), any()).thenReturn(None)
        when(core).wait(any()).thenReturn(None)

        # play our test sound file
        timing = play_sound(sound_file_path)

        # assert the response is as expected
        self.assertIsInstance(timing, list)

        # verify all the expected calls happended and the expected number of times
        verify(sf, times=1).read(sound_file_path, dtype='float32')
        verify(sd, times=1).play('data', MOCK_FS)
        verify(core, times=2).wait(any())

    def test_play_sound_raises_exception_if_soundfile_cannot_read_file(self):
        # fake sound file path
        sound_file_path = 'test_sound_file_path'

        # mock the other library interactions
        when(sf).read(
            sound_file_path, dtype='float32').thenRaise(Exception(''))

        # assert it raises the exception
        with self.assertRaises(Exception):
            play_sound(sound_file_path)

        # verify all the expected calls happended and the expected number of times
        verify(sf, times=1).read(sound_file_path, dtype='float32')

    def test_play_sound_sound_callback_evokes_with_timing(self):
        # fake sound file path
        sound_file_path = 'test_sound_file_path'
        test_trigger_name = 'test_trigger_name'
        test_trigger_time = 111
        self.test_timing = [test_trigger_name, test_trigger_time]

        experiment_clock = mock()

        def mock_callback_function(timing):
            self.assertEqual(timing, self.test_timing)

        # mock the other library interactions
        when(sf).read(
            sound_file_path, dtype='float32').thenReturn(('data', MOCK_FS))
        when(sd).play(any(), any()).thenReturn(None)
        when(core).wait(any()).thenReturn(None)
        when(experiment_clock).getTime().thenReturn(test_trigger_time)

        play_sound(
            sound_file_path,
            track_timing=True,
            sound_callback=mock_callback_function,
            trigger_name=test_trigger_name,
            experiment_clock=experiment_clock,
        )

        # verify all the expected calls happended and the expected number of times
        verify(sf, times=1).read(sound_file_path, dtype='float32')
        verify(sd, times=1).play('data', MOCK_FS)
        verify(core, times=2).wait(any())

    def test_soundfiles_generator(self):
        """Test that soundfiles function returns an cyclic generator."""

        directory = path.join('.', 'sounds')
        soundfile_paths = [
            path.join(directory, '0.wav'),
            path.join(directory, '1.wav'),
            path.join(directory, '2.wav')
        ]
        when(glob).glob(path.join(directory, '*.wav')).thenReturn(soundfile_paths)
        when(path).isdir(directory).thenReturn(True)

        gen = soundfiles(directory)
        self.assertEqual(next(gen), soundfile_paths[0])
        self.assertEqual(next(gen), soundfile_paths[1])
        self.assertEqual(next(gen), soundfile_paths[2])
        self.assertEqual(next(gen), soundfile_paths[0])
        self.assertEqual(next(gen), soundfile_paths[1])
        for _ in range(10):
            self.assertTrue(next(gen) in soundfile_paths)

    def test_soundfiles_generator_path_arg(self):
        """Test that soundfiles function constructs the correct path."""
        directory = path.join('.', 'sounds')
        soundfile_paths = [
            path.join(directory, '0.wav'),
            path.join(directory, '1.wav'),
            path.join(directory, '2.wav')
        ]
        when(glob).glob(path.join(directory, '*.wav')).thenReturn(soundfile_paths)
        when(path).isdir(directory).thenReturn(True)
        gen = soundfiles(directory)
        self.assertEqual(next(gen), soundfile_paths[0])

    def test_random_sequence_gen(self):
        """Test generation of random sequences"""
        alp = [
            'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
            'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
            '<', '_'
        ]
        stim_number = 10
        stim_length = 10
        seqs, seq_timings, seq_colors = random_rsvp_calibration_seq_gen(
            alp,
            timing=[0.5, 1, 0.2],
            color=['green', 'red', 'white'],
            stim_number=stim_number,
            stim_length=stim_length,
            is_txt=True)

        self.assertEqual(
            len(seqs), stim_number,
            'Should have produced the correct number of sequences')
        self.assertEqual(len(seq_timings), stim_number)
        self.assertEqual(len(seq_colors), stim_number)

        seq_strings = []
        for seq in seqs:
            self.assertEqual(
                len(seq), stim_length + 2,
                ('Sequence should include the correct number of choices as ',
                 'well as the target and cross.'))
            choices = seq[2:]
            self.assertEqual(stim_length, len(set(choices)),
                             'All choices should be unique')

            # create a string of the options
            seq_strings.append(''.join(choices))

        self.assertEqual(
            len(seqs), len(set(seq_strings)),
            'All sequences should be different')


    def test_best_selection(self):
        """Test best_selection"""

        self.assertEqual(['a', 'c', 'e'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.3, 0.1, 0.3, 0.1, 0.2],
                             len_query=3))

        # Test equal probabilities
        self.assertEqual(['a', 'b', 'c'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.2, 0.2, 0.2, 0.2, 0.2],
                             len_query=3))

        # Test always included
        self.assertEqual(
            ['a', 'c', 'd'],
            best_selection(
                selection_elements=['a', 'b', 'c', 'd', 'e'],
                val=[0.3, 0.1, 0.3, 0.1, 0.2],
                len_query=3,
                always_included=['d']),
            'Included item should bump out the best item with the lowest val.')

        self.assertEqual(
            ['a', 'b', 'c'],
            best_selection(
                selection_elements=['a', 'b', 'c', 'd', 'e'],
                val=[0.5, 0.4, 0.1, 0.0, 0.0],
                len_query=3,
                always_included=['b']),
            'Included item should retain its position if already present')

        self.assertEqual(['a', 'b', 'e'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.5, 0.0, 0.1, 0.3, 0.0],
                             len_query=3,
                             always_included=['b', 'e']),
                         'multiple included items should be supported.')

        self.assertEqual(['d'],
                         best_selection(
                             selection_elements=['a', 'b', 'c', 'd', 'e'],
                             val=[0.5, 0.4, 0.1, 0.0, 0.0],
                             len_query=1,
                             always_included=['d', 'e']),
                         'len_query should be respected.')

        self.assertEqual(['a', 'b', 'c'],
                    best_selection(
                        selection_elements=['a', 'b', 'c', 'd', 'e'],
                        val=[0.5, 0.4, 0.1, 0.0, 0.0],
                        len_query=3,
                        always_included=['<']),
                    'should ignore items not in the set.')

    def test_best_case_sequence_gen(self):
        """Test best_case_rsvp_seq_gen"""
        alp = ['a', 'b', 'c', 'd', 'e', 'f', 'g']
        n = 5
        samples, times, colors = best_case_rsvp_seq_gen(
            alp=alp,
            session_stimuli=[0.1, 0.1, 0.1, 0.2, 0.2, 0.1, 0.2],
            timing=[1, 0.2],
            color=['red', 'white'],
            stim_number=1,
            stim_length=n,
            is_txt=True)

        first_seq = samples[0]
        self.assertEqual(1, len(samples))
        self.assertEqual(n + 1, len(first_seq),
                         'Should include fixation cross.')
        self.assertEqual(len(samples), len(times))
        self.assertEqual(len(samples), len(colors))

        expected = ['+', 'a', 'b', 'd', 'e', 'g']
        for letter in expected:
            self.assertTrue(letter in first_seq)

        self.assertNotEqual(expected, first_seq, 'Should be in random order.')
        self.assertEqual([1] + ([0.2] * n), times[0])
        self.assertEqual(['red'] + (['white'] * n), colors[0])

    def test_best_case_sequence_gen_with_seq_constants(self):
        """Test best_case_rsvp_seq_gen with sequence constants"""

        alp = ['a', 'b', 'c', 'd', 'e', 'f', 'g']
        n = 5

        with self.assertRaises(
                Exception, msg='Constants should be in the alphabet'):
            best_case_rsvp_seq_gen(
                alp=alp,
                session_stimuli=[0.1, 0.1, 0.1, 0.2, 0.2, 0.1, 0.2],
                seq_constants=['<'])

        alp = ['a', 'b', 'c', 'd', 'e', 'f', 'g', '<']
        samples, times, colors = best_case_rsvp_seq_gen(
            alp=alp,
            session_stimuli=[0.1, 0.1, 0.1, 0.2, 0.2, 0.1, 0.2, 0.0],
            stim_number=1,
            stim_length=n,
            is_txt=True,
            seq_constants=['<'])

        first_seq = samples[0]
        self.assertEqual(1, len(samples))
        self.assertEqual(n + 1, len(first_seq),
                         'Should include fixation cross.')
        self.assertEqual(len(samples), len(times))
        self.assertEqual(len(samples), len(colors))

        expected = ['+', 'a', 'd', 'e', 'g', '<']
        for letter in expected:
            self.assertTrue(letter in first_seq)

        self.assertNotEqual(expected, first_seq, 'Should be in random order.')
        self.assertEqual([1] + ([0.2] * n), times[0])
        self.assertEqual(['red'] + (['white'] * n), colors[0])


if __name__ == '__main__':
    unittest.main()



import pickle
import logging

from bcipy.helpers.load import (
    read_data_csv,
    load_experimental_data,
    load_json_parameters)
from bcipy.signal.process.filter import bandpass, notch, downsample
from bcipy.signal.model.mach_learning.train_model import train_pca_rda_kde_model
from bcipy.helpers.task import trial_reshaper
from bcipy.helpers.vizualization import generate_offline_analysis_screen
from bcipy.helpers.triggers import trigger_decoder
from bcipy.helpers.acquisition import analysis_channels,\
    analysis_channel_names_by_pos
from bcipy.helpers.stimuli import play_sound

log = logging.getLogger(__name__)


def offline_analysis(data_folder: str=None, parameters: dict={}, alert_finished: bool=True):
    """ Gets calibration data and trains the model in an offline fashion.
        pickle dumps the model into a .pkl folder
        Args:
            data_folder(str): folder of the data
                save all information and load all from this folder
            parameter(dict): parameters for running offline analysis
            alert_finished(bool): whether or not to alert the user offline analysis complete

        How it Works:
        - reads data and information from a .csv calibration file
        - reads trigger information from a .txt trigger file
        - filters data
        - reshapes and labels the data for the training procedure
        - fits the model to the data
            - uses cross validation to select parameters
            - based on the parameters, trains system using all the data
        - pickle dumps model into .pkl file
        - generates and saves offline analysis screen
        - [optional] alert the user finished processing
    """

    if not data_folder:
        data_folder = load_experimental_data()

    mode = 'calibration'
    trial_length = parameters.get('collection_window_after_trial_length')

    raw_dat, _, channels, type_amp, fs = read_data_csv(
        data_folder + '/' + parameters.get('raw_data_name', 'raw_data.csv'))

    log.info(f'Channels read from csv: {channels}')
    log.info(f'Device type: {type_amp}')

    downsample_rate = parameters.get('down_sampling_rate', 2)

    # Remove 60hz noise with a notch filter
    notch_filter_data = notch.notch_filter(raw_dat, fs, frequency_to_remove=60)

    # bandpass filter from 2-45hz
    filtered_data = bandpass.butter_bandpass_filter(notch_filter_data, 2, 45, fs, order=2)

    # downsample
    data = downsample.downsample(filtered_data, factor=downsample_rate)

    # Process triggers.txt
    triggers_file = parameters.get('trigger_file_name', 'triggers.txt')
    _, t_t_i, t_i, offset = trigger_decoder(
        mode=mode,
        trigger_path=f'{data_folder}/{triggers_file}')

    static_offset = parameters.get('static_trigger_offset', 0)

    offset = offset + static_offset

    # Channel map can be checked from raw_data.csv file.
    # read_data_csv already removes the timespamp column.
    channel_map = analysis_channels(channels, type_amp)

    x, y, _, _ = trial_reshaper(t_t_i, t_i, data,
                                      mode=mode, fs=fs, k=downsample_rate,
                                      offset=offset,
                                      channel_map=channel_map,
                                      trial_length=trial_length)

    k_folds = parameters.get('k_folds', 10)
    model, auc = train_pca_rda_kde_model(x, y, k_folds=k_folds)

    log.info('Saving offline analysis plots!')

    # After obtaining the model get the transformed data for plotting purposes
    model.transform(x)
    generate_offline_analysis_screen(
        x, y, model=model, folder=data_folder,
        down_sample_rate=downsample_rate,
        fs=fs, save_figure=True, show_figure=False,
        channel_names = analysis_channel_names_by_pos(channels, channel_map))

    log.info('Saving the model!')
    with open(data_folder + f'/model_{auc}.pkl', 'wb') as output:
        pickle.dump(model, output)

    if alert_finished:
        offline_analysis_tone = parameters.get('offline_analysis_tone')
        play_sound(offline_analysis_tone)

    return model


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--data_folder', default=None)
    parser.add_argument('-p', '--parameters_file',
                        default='bcipy/parameters/parameters.json')
    args = parser.parse_args()

    print(f'Loading params from {args.parameters_file}')
    parameters = load_json_parameters(args.parameters_file,
                                      value_cast=True)
    offline_analysis(args.data_folder, parameters)



from psychopy import visual, clock
import numpy as np
import time
from bcipy.helpers.task import alphabet

import logging


class MatrixDisplay:
    """ Matrix Display Object for Sequence Presentation. Animates a sequence
        in Matrix. Mode should be determined outside.
        Attr:
            task(visual_Text_Stimuli): task bar
            text(list[visual_Text_Stimuli]): text bar, there can be
                different number of information texts in different paradigms
            sti(visual_Text_Stimuli): stimuli text
            trialClock(core_clock): timer for presentation """

    def __init__(self, window, clock, experiment_clock, task_color='white',
                 task_font='Times', task_pos=(-.75, .75), task_height=0.1,
                 grid_rows=6, grid_columns=6, time_flash=.25,
                 task_text='1/100',
                 stim_font='Times',
                 stim_pos=(0, .0),
                 stim_height=0.25,
                 is_txt_stim=True, alp=None):

        self.win = window
        self.logger = log = logging.getLogger(__name__)

        # TASK TEXT
        self.task_font_text = task_font
        self.task_height = task_height
        self.task_pos = task_pos
        self.task_text = task_text
        self.task_color = task_color

        # STIM / GRID
        self.is_txt_stim = is_txt_stim
        self.stimuli = []
        self.rows = grid_rows
        self.stimuli_font = stim_font
        self.columns = grid_columns
        self.stim_height = stim_height

        self.stim_number = grid_rows + grid_columns

        self.max_height_grid = -1 + stim_height
        self.max_width_grid = 1 - stim_height
        self.uniform_grid_values_row = sorted(
            np.linspace(
                start=self.max_height_grid,
                stop=self.max_width_grid,
                num=self.rows))

        self.uniform_grid_values_col = sorted(
            np.linspace(
                start=self.max_height_grid,
                stop=self.max_width_grid - .25,
                num=self.columns),
            reverse=True)

        if not alp:
            self.alp = alphabet()
        else:
            self.alp = alp

        self.flash = time_flash
        # Clocks
        self.trialClock = clock
        self.expClock = experiment_clock

        # Length of the stimuli (number of stimuli on screen)
        self.stim_length = len(self.stimuli)

    def draw_static(self):
        """ Draws static elements in a stimulus. """

        # Draw the grid
        if len(self.stimuli) is 0:
            self.make_spelling_grid()

        for stim in self.stimuli:
            stim.draw()

        # Next, Draw the task text
        task_text = visual.TextStim(win=self.win, color=self.task_color,
                                    height=self.task_height,
                                    text=self.task_text,
                                    font=self.task_font_text,
                                    pos=self.task_pos,
                                    wrapWidth=None, colorSpace='rgb',
                                    opacity=1, depth=-6.0)
        task_text.draw()

        self.win.flip()

    def make_spelling_grid(self):
        # First Draw the grid!
        alp_idx = 0

        # Loop through each row
        for idx in range(self.rows):
            if self.is_txt_stim:

                # loop through each column
                for x in range(self.columns):
                    if alp_idx > len(self.alp) - 1:
                        break

                    pos = determine_position_on_grid(
                        idx, x,
                        self.uniform_grid_values_col,
                        self.uniform_grid_values_row)

                    stim = visual.TextStim(win=self.win,
                                           height=self.stim_height,
                                           text=self.alp[alp_idx],
                                           font=self.stimuli_font, pos=pos,
                                           wrapWidth=None, colorSpace='rgb',
                                           opacity=1, depth=-6.0)
                    alp_idx += 1

                    stim.draw()
                    self.stimuli.append(stim)

    def do_sequence(self):
        """ implements grid sequence """

        # Stimulate rows
        for x in self.uniform_grid_values_row:
            for stim in self.stimuli:
                if stim.pos[0] == x:
                    stim.color = 'blue'
                else:
                    stim.color = 'white'

            self.draw_static()
            time.sleep(self.flash)

        # Stimulate columns
        for x in self.uniform_grid_values_col:
            for stim in self.stimuli:
                if stim.pos[1] == x:
                    stim.color = 'blue'
                else:
                    stim.color = 'white'

            self.draw_static()
            time.sleep(self.flash)


def determine_position_on_grid(row_idx, col_idx,
                               uniform_grid_values_row,
                               uniform_grid_values_col):
    # Position on screen goes from -1 to 1, where (0, 0) is the center of the
    #  screen. Assume an adequate size ratio of window.
    try:
        x = uniform_grid_values_col[col_idx]
        y = uniform_grid_values_row[row_idx]
    except Exception as e:
        log.debug(f'at index: {col_idx} Error: {e}')

    # # else:
    # x = 0

    return (x, y)

if __name__ == "__main__":
    # How many stimulations?
    task_length = 5

    # make the display window
    display_window = visual.Window(
        size=[1000,
              1000],
        screen=0,
        allowGUI=False,
        useFBO=False,
        fullscr=False,
        allowStencil=False,
        monitor='mainMonitor',
        winType='pyglet', units='norm', waitBlanking=True,
        color='black')

    # Create a Display Matrix Object
    matrix = MatrixDisplay(
        display_window,
        clock.MonotonicClock(),
        clock.MonotonicClock(),
        task_text='')

    # draw matrix grid and other static
    matrix.draw_static()

    task_index = 0
    for i in range(task_length):
        task_index += 1

        matrix.task_text = '%s/5' % task_index
        matrix.draw_static()

        # animate!
        matrix.do_sequence()

        time.sleep(2)



from typing import Dict, List
import numpy as np
import bcipy.acquisition.datastream.generator as generator
import bcipy.acquisition.protocols.registry as registry
from bcipy.acquisition.client import DataAcquisitionClient, CountClock
from bcipy.acquisition.datastream.server import start_socket_server, await_start
from bcipy.acquisition.processor import FileWriter, DispatchProcessor
from bcipy.acquisition.datastream.lsl_server import LslDataServer
from bcipy.gui.viewer.processor.viewer_processor import ViewerProcessor

# Channels relevant for analysis, for each currently supported device.
#  Note this leaves out triggers or other non-eeg channels. If desired,
#   they should be added to this list.
analysis_channels_by_device = {
    'DSI': ["P3", "C3", "F3", "Fz", "F4", "C4", "P4", "Cz", "A1", "Fp1", "Fp2",
            "T3", "T5", "O1", "O2", "F7", "F8", "A2", "T6", "T4"],
    'DSI_VR300': ["P4", "Fz", "Pz", "F7", "PO8", "PO7", "Oz"],
    'g.USBamp-2': ["Ch1", "Ch2", "Ch3", "Ch4", "Ch5", "Ch6", "Ch7", "Ch8",
                   "Ch9", "Ch10", "Ch11", "Ch12", "Ch13", "Ch14", "Ch15",
                   "Ch16"],
    'LSL': ["ch1", "ch2", "ch3", "ch4", "ch5", "ch6", "ch7", "ch8",
            "ch9", "ch10", "ch11", "ch12", "ch13", "ch14", "ch15", "ch16"]
}


def init_eeg_acquisition(parameters: dict, save_folder: str,
                         clock=CountClock(), server: bool=False):
    """Initialize EEG Acquisition.

    Initializes a client that connects with the EEG data source and begins
    data collection.

    Parameters
    ----------
        parameters : dict
            configuration details regarding the device type and other relevant
            connection information.
             {
               "acq_device": str,
               "acq_host": str,
               "acq_port": int,
               "buffer_name": str,
               "raw_data_name": str
             }
        clock : Clock, optional
            optional clock used in the client; see client for details.
        server : bool, optional
            optionally start a server that streams random DSI data; defaults
            to true; if this is True, the client will also be a DSI client.
    Returns
    -------
        (client, server) tuple
    """

    # Initialize the needed DAQ Parameters
    host = parameters['acq_host']
    port = parameters['acq_port']

    parameters = {
        'acq_show_viewer': parameters['acq_show_viewer'],
        'viewer_screen': 1 if int(parameters['stim_screen']) == 0 else 0,
        'buffer_name': save_folder + '/' + parameters['buffer_name'],
        'device': parameters['acq_device'],
        'filename': save_folder + '/' + parameters['raw_data_name'],
        'connection_params': {'host': host,
                              'port': port}}

    # Set configuration parameters (with default values if not provided).
    buffer_name = parameters.get('buffer_name', 'buffer.db')
    connection_params = parameters.get('connection_params', {})
    device_name = parameters.get('device', 'DSI')
    filename = parameters.get('filename', 'rawdata.csv')

    dataserver = False
    if server:
        if device_name == 'DSI':
            protocol = registry.default_protocol(device_name)
            dataserver, port = start_socket_server(protocol, host, port)
            connection_params['port'] = port
        elif device_name == 'LSL':
            channel_count = 16
            sample_rate = 256
            channels = ['ch{}'.format(c + 1) for c in range(channel_count)]
            dataserver = LslDataServer(params={'name': 'LSL',
                                               'channels': channels,
                                               'hz': sample_rate},
                                       generator=generator.random_data(
                                           channel_count=channel_count))
            await_start(dataserver)
        else:
            raise ValueError(
                'Server (fake data mode) for this device type not supported')

    Device = registry.find_device(device_name)

    filewriter = FileWriter(filename=filename)
    proc = filewriter
    if parameters['acq_show_viewer']:
        proc = DispatchProcessor(
            filewriter,
            ViewerProcessor(display_screen=parameters['viewer_screen']))

    # Start a client. We assume that the channels and fs will be set on the
    # device; add a channel parameter to Device to override!
    client = DataAcquisitionClient(
        device=Device(connection_params=connection_params),
        processor=proc,
        buffer_name=buffer_name,
        clock=clock)

    client.start_acquisition()

    # If we're using a server or data generator, there is no reason to
    # calibrate data.
    if server and device_name != 'LSL':
        client.is_calibrated = True

    return (client, dataserver)


def analysis_channels(channels: List[str], device_name: str) -> list:
    """Analysis Channels.

    Defines the channels within a device that should be used for analysis.

    Parameters:
    ----------
        channels(list(str)): list of channel names from the raw_data
            (excluding the timestamp)
        device_name(str): daq_type from the raw_data file.
    Returns:
    --------
        A binary list indicating which channels should be used for analysis.
        If i'th element is 0, i'th channel in filtered_eeg is removed.
    """
    relevant_channels = analysis_channels_by_device.get(device_name)
    if not relevant_channels:
        raise Exception("Analysis channels for the given device not found: "
                        f"{device_name}.")
    if channels is None:
        return relevant_channels
    return [int(ch in relevant_channels) for ch in channels]


def analysis_channel_names_by_pos(channels: List[str],
                                  channel_map: List[int]) -> Dict[int, str]:
    """Generate a dict mapping the index of a channel (used in offline
    analysis) to its name.

    Parameters:
    -----------
        channels - list of channel names
        channel_map - output of analysis_channels method; binary list
            where each item is either 0 or 1.
    """
    selection = [bool(x) for x in channel_map]
    selected_channels = np.array(channels)[selection]
    return {i: ch for i, ch in enumerate(selected_channels)}
